'''
http.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.

The CaseInsensitiveDict implementation is copyright 2013 Kenneth Reitz and
licensed under the Apache License, Version 2.0
(http://www.apache.org/licenses/LICENSE-2.0)
'''

import email
import email.policy
import errno
import hashlib
import logging
import os
import select
import socket
import ssl
import textwrap
from base64 import b64encode
from collections import deque
from collections.abc import Mapping, MutableMapping
from enum import Enum
from http.client import HTTP_PORT, HTTPS_PORT, NO_CONTENT, NOT_MODIFIED
from inspect import getdoc
from typing import Union

import trio

log = logging.getLogger(__name__)

#: Internal buffer size
BUFFER_SIZE = 64 * 1024

#: Maximal length of HTTP status line. If the server sends a line longer than
#: this value, `InvalidResponse` will be raised.
MAX_LINE_SIZE = BUFFER_SIZE - 1

#: Maximal length of a response header (i.e., for all header
#: lines together). If the server sends a header segment longer than
#: this value, `InvalidResponse` will be raised.
MAX_HEADER_SIZE = BUFFER_SIZE - 1


class Symbol:
    '''
    A symbol instance represents a specific state. Its value is
    not relevant, as it should only ever be assigned to or compared
    with other variables.
    '''

    __slots__ = ['name']

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return f'Symbol({self.name})'


class Encodings(Enum):
    CHUNKED = 1
    IDENTITY = 2


#: Sentinel for `HTTPConnection._out_remaining` to indicate that
#: we're waiting for a 100-continue response from the server
WAITING_FOR_100c = Symbol('WAITING_FOR_100c')

#: Sentinel for `HTTPConnection._in_remaining` to indicate that
#: the response body cannot be read and that an exception
#: (stored in the `HTTPConnection._encoding` attribute) should
#: be raised.
RESPONSE_BODY_ERROR = Symbol('RESPONSE_BODY_ERROR')

#: Sentinel for `HTTPConnection._in_remaining` to indicate that
#: we should read until EOF (i.e., no keep-alive)
READ_UNTIL_EOF = Symbol('READ_UNTIL_EOF')

#: Sequence of ``(hostname, port)`` tuples that are used by distinguish between permanent and
#: temporary name resolution problems.
DNS_TEST_HOSTNAMES = (('www.google.com', 80), ('www.iana.org', 80), ('C.root-servers.org', 53))


class HTTPResponse:
    '''
    This class encapsulates information about HTTP response.  Instances of this
    class are returned by the `HTTPConnection.read_response` method and have
    access to response status, reason, and headers.  Response body data
    has to be read directly from the `HTTPConnection` instance.
    '''

    def __init__(self, method, path, status, reason, headers, length=None):
        #: HTTP Method of the request this was response is associated with
        self.method = method

        #: Path of the request this was response is associated with
        self.path = path

        #: HTTP status code returned by the server
        self.status = status

        #: HTTP reason phrase returned by the server
        self.reason = reason

        #: HTTP Response headers, a `email.message.Message` instance
        self.headers = headers

        #: Length of the response body or `None` if not known. This attribute
        #: contains the actual length of the *transmitted* response. That means
        #: that for responses where RFC 2616 mandates that no request body
        #: be sent (e.g. in response to HEAD requests or for 1xx response
        #: codes) this value is zero. In these cases, the length of the body that
        #: *would* have been send can be extracted from the ``Content-Length``
        #: response header.
        self.length = length


class BodyFollowing:
    '''
    Sentinel class for the *body* parameter of the
    `~HTTPConnection.send_request` method. Passing an instance of this class
    declares that body data is going to be provided in separate method calls.

    If no length is specified in the constructor, the body data will be send
    using chunked encoding.
    '''

    __slots__ = 'length'

    def __init__(self, length=None):
        #: the length of the body data that is going to be send, or `None`
        #: to use chunked encoding.
        self.length = length


class _ChunkTooLong(Exception):
    '''
    Raised by `_co_readstr_until` if the requested end pattern
    cannot be found within the specified byte limit.
    '''

    pass


class _GeneralError(Exception):
    msg = 'General HTTP Error'

    def __init__(self, msg=None):
        if msg:
            self.msg = msg

    def __str__(self):
        return self.msg


class HostnameNotResolvable(Exception):
    '''Raised if a host name does not resolve to an ip address.

    This exception is raised if a resolution attempt results in a `socket.gaierror` or
    `socket.herror` exception with errno :const:`!socket.EAI_NODATA` or
    :const:`!socket.EAI_NONAME`, but at least one of the hostnames in `DNS_TEST_HOSTNAMES`
    can be resolved.
    '''

    def __init__(self, hostname):
        self.name = hostname

    def __str__(self):
        return 'Host %s does not have any ip addresses' % self.name


class DNSUnavailable(Exception):
    '''Raised if the DNS server cannot be reached.

    This exception is raised if a resolution attempt results in a `socket.gaierror` or
    `socket.herror` exception with errno :const:`socket.EAI_AGAIN`, or if none of the
    hostnames in `DNS_TEST_HOSTNAMES` can be resolved.
    '''

    def __init__(self, hostname):
        self.name = hostname

    def __str__(self):
        return 'Unable to resolve %s, DNS server unavailable.' % self.name


class HostnameNotResolvableOrDNSUnavailable(Exception):
    '''Raised if a host name does not resolve or DNS server cannot be reached.'''

    def __init__(self, hostname):
        self.name = hostname

    def __str__(self):
        return 'Unable to resolve host %s' % self.name


class StateError(_GeneralError):
    '''
    Raised when attempting an operation that doesn't make
    sense in the current connection state.
    '''

    msg = 'Operation invalid in current connection state'


class ExcessBodyData(_GeneralError):
    '''
    Raised when trying to send more data to the server than
    announced.
    '''

    msg = 'Cannot send larger request body than announced'


class InvalidResponse(_GeneralError):
    '''
    Raised if the server produced an invalid response (i.e, something
    that is not proper HTTP 1.0 or 1.1).
    '''

    msg = 'Server sent invalid response'


class UnsupportedResponse(_GeneralError):
    '''
    This exception is raised if the server produced a response that is not
    supported. This should not happen for servers that are HTTP 1.1 compatible.

    If an `UnsupportedResponse` exception has been raised, this typically means
    that synchronization with the server will be lost (i.e., we cannot
    determine where the current response ends and the next response starts), so
    the connection needs to be reset by calling the
    :meth:`~HTTPConnection.disconnect` method.
    '''

    msg = 'Server sent unsupported response'


class ConnectionClosed(_GeneralError):
    '''
    Raised if the connection was unexpectedly closed.

    This exception is raised also if the server declared that it will close the
    connection (by sending a ``Connection: close`` header). Such responses can
    still be read completely, but the next attempt to send a request or read a
    response will raise the exception. To re-use the connection after the server
    has closed the connection, call `HTTPConnection.reset` before further
    requests are send.

    This behavior is intentional, because the caller may have already issued
    other requests (i.e., used pipelining). By raising an exception, the caller
    is notified that any pending requests have been lost and need to be resend.
    '''

    msg = 'connection closed unexpectedly'


class ConnectionTimedOut(_GeneralError):
    '''
    Raised if a regular `HTTPConnection` method (i.e., no coroutine) was
    unable to send or receive data for the timeout specified in the
    `HTTPConnection.timeout` attribute.
    '''

    msg = 'send/recv timeout exceeded'


class _Buffer:
    '''
    This class represents a buffer with a fixed size, but varying
    fill level.
    '''

    __slots__ = ('d', 'b', 'e')

    def __init__(self, size):
        #: Holds the actual data
        self.d = bytearray(size)

        #: Position of the first buffered byte that has not yet
        #: been consumed ("*b*eginning")
        self.b = 0

        #: Fill-level of the buffer (e is for "end")
        self.e = 0

    def __len__(self):
        '''Return amount of data ready for consumption'''
        return self.e - self.b

    def clear(self):
        '''Forget all buffered data'''

        self.b = 0
        self.e = 0

    def compact(self):
        '''Ensure that buffer can be filled up to its maximum size

        If part of the buffer data has been consumed, the unconsumed part is
        copied to the beginning of the buffer to maximize the available space.
        '''

        if self.b == 0:
            return

        log.debug('compacting buffer')
        buf = memoryview(self.d)[self.b : self.e]
        len_ = len(buf)
        self.d = bytearray(len(self.d))
        self.d[:len_] = buf
        self.b = 0
        self.e = len_

    def exhaust(self):
        '''Return (and consume) all available data'''

        if self.b == 0:
            log.debug('exhausting buffer (truncating)')
            # Return existing buffer after truncating it
            buf = self.d
            self.d = bytearray(len(self.d))
            buf[self.e :] = b''
        else:
            log.debug('exhausting buffer (copying)')
            buf = self.d[self.b : self.e]

        self.b = 0
        self.e = 0

        return buf


class HTTPConnection:
    '''
    This class encapsulates a HTTP connection.

    Methods whose name begin with ``co_`` return coroutines. Instead of
    blocking, a coroutines will yield a `PollNeeded` instance that encapsulates
    information about the IO operation that would block. The coroutine should be
    resumed once the operation can be performed without blocking.

    `HTTPConnection` instances can be used as context managers. The
    `.disconnect` method will be called on exit from the managed block.
    '''

    def __init__(self, hostname, port=None, ssl_context=None, proxy=None):
        if port is None:
            if ssl_context is None:
                self.port = HTTP_PORT
            else:
                self.port = HTTPS_PORT
        else:
            self.port = port

        self.ssl_context = ssl_context
        self.hostname = hostname

        #: Socket object connecting to the server
        self._sock = None

        #: `select.poll` method to query if connection is ready for sending
        self._poll_send = None

        #: `select.poll` method to query if connection is ready for receiving
        self._poll_recv = None

        #: Read-buffer
        self._rbuf = _Buffer(BUFFER_SIZE)

        #: a tuple ``(hostname, port)`` of the proxy server to use or `None`.
        self.proxy = proxy

        #: a deque of ``(method, path, body_len)`` tuples corresponding to
        #: requests whose response has not yet been read completely. Requests
        #: with Expect: 100-continue will be added twice to this queue, once
        #: after the request header has been sent, and once after the request
        #: body data has been sent. *body_len* is `None`, or the size of the
        #: **request** body that still has to be sent when using 100-continue.
        self._pending_requests = deque()

        #: This attribute is `None` when a request has been sent completely.  If
        #: request headers have been sent, but request body data is still
        #: pending, it is set to a ``(method, path, body_len)`` tuple. *body_len*
        #: is the number of bytes that that still need to send, or
        #: `WAITING_FOR_100c` if we are waiting for a 100 response from the server.
        self._out_remaining = None

        #: Number of remaining bytes of the current response body (or current
        #: chunk), `None` if there is no active response or `READ_UNTIL_EOF` if
        #: we have to read until the connection is closed (i.e., we don't know
        #: the content-length and keep-alive is not active).
        self._in_remaining = None

        #: Transfer encoding of the active response (if any).
        self._encoding = None

        #: If a regular `HTTPConnection` method is unable to send or receive data for more than
        #: this period (in ms), it will raise `ConnectionTimedOut`.
        self._timeout_ms: int = 24 * 60 * 60 * 1000

        #: Filehandler for tracing
        self.trace_fh = None

        #: If set, coroutines will never yield and instead block internally.
        #: (This is used for implementing the blocking/non-coroutine API).
        self._sync_context = False

    @property
    def timeout(self) -> float:
        return self._timeout_ms / 1000

    @timeout.setter
    def timeout(self, value: Union[int, float, None]) -> None:
        if value is None:
            self._timeout_ms = 24 * 60 * 60 * 1000
        else:
            self._timeout_ms = int(value * 1000)

    # Implement bare-bones `io.BaseIO` interface, so that instances
    # can be wrapped in `io.TextIOWrapper` if desired.
    def writable(self):
        return True

    def readable(self):
        return True

    def seekable(self):
        return False

    # One could argue that the stream should be considered closed if
    # there is no active response. However, this breaks TextIOWrapper
    # (which fails if the stream becomes closed even after b'' has
    # been read), so we just declare to be always open.
    closed = False

    def connect(self):
        """Connect to the remote server

        This method generally does not need to be called manually.
        """

        log.debug('start')

        if self.proxy:
            log.debug('connecting to %s', self.proxy)
            self._sock = create_socket(self.proxy)
            if self.ssl_context:
                self._tunnel()
        else:
            log.debug('connecting to %s', (self.hostname, self.port))
            self._sock = create_socket((self.hostname, self.port))

        if self.ssl_context:
            log.debug('establishing ssl layer')

            self._sock = self.ssl_context.wrap_socket(self._sock, server_hostname=self.hostname)

        self._sock.setblocking(False)
        self._rbuf.clear()
        self._out_remaining = None
        self._in_remaining = None
        self._pending_requests = deque()

        poll = select.poll()
        poll.register(self._sock, select.POLLIN)
        self._poll_recv = poll.poll

        poll = select.poll()
        poll.register(self._sock, select.POLLOUT)
        self._poll_send = poll.poll

        if 'S3QL_HTTP_TRACEFILE' in os.environ:
            self.trace_fh = open(
                os.environ['S3QL_HTTP_TRACEFILE'] % id(self._sock), 'wb+', buffering=0
            )

        log.debug('done')

    def _tunnel(self):
        '''placeholder, will be replaced dynamically'''
        orig_context = self._sync_context
        self._sync_context = True
        try:
            return eval_coroutine(self._co_tunnel())
        finally:
            self._sync_context = orig_context

    async def _co_tunnel(self):
        '''Set up CONNECT tunnel to destination server'''

        log.debug('start connecting to %s:%d', self.hostname, self.port)

        await self._co_send(
            ("CONNECT %s:%d HTTP/1.0\r\n\r\n" % (self.hostname, self.port)).encode('latin1')
        )

        (status, reason) = await self._co_read_status()
        log.debug('got %03d %s', status, reason)
        await self._co_read_header()

        if status != 200:
            self.disconnect()
            raise ConnectionError("Tunnel connection failed: %d %s" % (status, reason))

    def send_request(self, method, path, headers=None, body=None, expect100=False):
        '''placeholder, will be replaced dynamically'''
        orig_context = self._sync_context
        self._sync_context = True
        try:
            return eval_coroutine(
                self.co_send_request(method, path, headers=headers, body=body, expect100=expect100)
            )
        finally:
            self._sync_context = orig_context

    async def co_send_request(self, method, path, headers=None, body=None, expect100=False):
        '''Send a new HTTP request to the server

        The message body may be passed in the *body* argument or be sent
        separately. In the former case, *body* must be a :term:`bytes-like
        object`. In the latter case, *body* must be an a `BodyFollowing`
        instance specifying the length of the data that will be sent. If no
        length is specified, the data will be send using chunked encoding.

        *headers* should be a mapping containing the HTTP headers to be send
        with the request. Multiple header lines with the same key are not
        supported. It is recommended to pass a `CaseInsensitiveDict` instance,
        other mappings will be converted to `CaseInsensitiveDict` automatically.

        If *body* is a provided as a :term:`bytes-like object`, a
        ``Content-MD5`` header is generated automatically unless it has been
        provided in *headers* already.
        '''

        log.debug('start')

        if expect100 and not isinstance(body, BodyFollowing):
            raise ValueError('expect100 only allowed for separate body')

        if self._sock is None:
            self.connect()

        if self._out_remaining:
            raise StateError('body data has not been sent completely yet')

        if headers is None:
            headers = CaseInsensitiveDict()
        elif not isinstance(headers, CaseInsensitiveDict):
            headers = CaseInsensitiveDict(headers)

        pending_body_size = None
        if body is None:
            headers['Content-Length'] = '0'
        elif isinstance(body, BodyFollowing):
            if body.length is None:
                raise ValueError('Chunked encoding not yet supported.')
            log.debug('preparing to send %d bytes of body data', body.length)
            if expect100:
                headers['Expect'] = '100-continue'
                # Do not set _out_remaining, we must only send data once we've
                # read the response. Instead, save body size in
                # _pending_requests so that it can be restored by
                # read_response().
                pending_body_size = body.length
                self._out_remaining = (method, path, WAITING_FOR_100c)
            else:
                self._out_remaining = (method, path, body.length)
            headers['Content-Length'] = str(body.length)
            body = None
        elif isinstance(body, (bytes, bytearray, memoryview)):
            headers['Content-Length'] = str(len(body))
            if 'Content-MD5' not in headers:
                log.debug('computing content-md5')
                headers['Content-MD5'] = b64encode(hashlib.md5(body).digest()).decode('ascii')
        else:
            raise TypeError('*body* must be None, bytes-like or BodyFollowing')

        # Generate host header
        host = self.hostname
        if host.find(':') >= 0:
            host = '[{}]'.format(host)
        default_port = HTTPS_PORT if self.ssl_context else HTTP_PORT
        if self.port == default_port:
            headers['Host'] = host
        else:
            headers['Host'] = '{}:{}'.format(host, self.port)

        # Assemble request
        headers['Accept-Encoding'] = 'identity'
        if 'Connection' not in headers:
            headers['Connection'] = 'keep-alive'
        if self.proxy and not self.ssl_context:
            gpath = "http://{}{}".format(headers['Host'], path)
        else:
            gpath = path
        request = ['{} {} HTTP/1.1'.format(method, gpath).encode('latin1')]
        for key, val in headers.items():
            request.append('{}: {}'.format(key, val).encode('latin1'))
        request.append(b'')

        if body is not None:
            request.append(body)
        else:
            request.append(b'')

        buf = b'\r\n'.join(request)

        log.debug('sending %s %s', method, path)
        await self._co_send(buf)
        if not self._out_remaining or expect100:
            self._pending_requests.append((method, path, pending_body_size))

    async def _co_send(self, buf):
        '''Send *buf* to server'''

        log.debug('trying to send %d bytes', len(buf))

        if not isinstance(buf, memoryview):
            buf = memoryview(buf)

        while True:
            try:
                if self._sock is None:
                    raise ConnectionClosed('connection has been closed locally')
                if self.trace_fh:
                    self.trace_fh.write(buf)
                len_ = self._sock.send(buf)
                # An SSL socket has the nasty habit of returning zero
                # instead of raising an exception when in non-blocking
                # mode.
                if len_ == 0:
                    raise BlockingIOError()
            except (socket.timeout, ssl.SSLWantWriteError, BlockingIOError):
                log.debug('yielding')
                if self._sync_context:
                    if not self._poll_send(self._timeout_ms):
                        raise ConnectionTimedOut('Timeout in send()')
                else:
                    with trio.move_on_after(self._timeout_ms / 1000) as ctx:
                        await trio.lowlevel.wait_writable(self._sock)
                    if ctx.cancelled_caught:
                        raise ConnectionTimedOut('Timeout in send()')
                continue
            except (BrokenPipeError, ConnectionResetError, ssl.SSLEOFError):
                raise ConnectionClosed('connection was interrupted')
            except OSError as exc:
                if exc.errno == errno.EINVAL:
                    # Blackhole routing, according to ip(7)
                    raise ConnectionClosed('ip route goes into black hole')
                else:
                    raise
            except InterruptedError:
                log.debug('interrupted')
                # According to send(2), this means that no data has been sent
                # at all before the interruption, so we just try again.
                continue

            log.debug('sent %d bytes', len_)
            buf = buf[len_:]
            if len(buf) == 0:
                log.debug('done')
                return

    def write(self, buf):
        '''placeholder, will be replaced dynamically'''
        orig_context = self._sync_context
        self._sync_context = True
        try:
            return eval_coroutine(self.co_write(buf))
        finally:
            self._sync_context = orig_context

    async def co_write(self, buf):
        '''Write request body data

        `ExcessBodyData` will be raised when attempting to send more data than
        required to complete the request body of the active request.
        '''

        log.debug('start (len=%d)', len(buf))

        if not self._out_remaining:
            raise StateError('No active request with pending body data')

        (method, path, remaining) = self._out_remaining
        if remaining is WAITING_FOR_100c:
            raise StateError("can't write when waiting for 100-continue")

        if len(buf) > remaining:
            raise ExcessBodyData(
                'trying to write %d bytes, but only %d bytes pending' % (len(buf), remaining)
            )

        try:
            await self._co_send(buf)
        except ConnectionClosed:
            # If the server closed the connection, we pretend that all data
            # has been sent, so that we can still read a (buffered) error
            # response.
            self._out_remaining = None
            self._pending_requests.append((method, path, None))
            raise

        len_ = len(buf)
        if len_ == remaining:
            log.debug('body sent fully')
            self._out_remaining = None
            self._pending_requests.append((method, path, None))
        else:
            self._out_remaining = (method, path, remaining - len_)

        log.debug('done')

    def response_pending(self):
        '''Return `True` if there are still outstanding responses

        This includes responses that have been partially read.
        '''

        return self._sock is not None and len(self._pending_requests) > 0

    def read_response(self):
        '''placeholder, will be replaced dynamically'''
        orig_context = self._sync_context
        self._sync_context = True
        try:
            return eval_coroutine(self.co_read_response())
        finally:
            self._sync_context = orig_context

    async def co_read_response(self):
        '''Read response status line and headers

        Return a `HTTPResponse` instance containing information about response
        status, reason, and headers. The response body data must be retrieved
        separately (e.g. using `.read` or `.readall`).
        '''

        log.debug('start')

        if len(self._pending_requests) == 0:
            raise StateError('No pending requests')

        if self._in_remaining is not None:
            raise StateError('Previous response not read completely')

        (method, path, body_size) = self._pending_requests[0]

        # Need to loop to handle any 1xx responses
        while True:
            (status, reason) = await self._co_read_status()
            log.debug('got %03d %s', status, reason)

            hstring = await self._co_read_header()
            header = email.message_from_string(hstring, policy=email.policy.HTTP)

            if status < 100 or status > 199:
                break

            # We are waiting for 100-continue
            if body_size is not None and status == 100:
                break

        # Handle (expected) 100-continue
        if status == 100:
            assert self._out_remaining == (method, path, WAITING_FOR_100c)

            # We're ready to sent request body now
            self._out_remaining = self._pending_requests.popleft()
            self._in_remaining = None

            # Return early, because we don't have to prepare
            # for reading the response body at this time
            return HTTPResponse(method, path, status, reason, header, length=0)

        # Handle non-100 status when waiting for 100-continue
        elif body_size is not None:
            assert self._out_remaining == (method, path, WAITING_FOR_100c)
            # RFC 2616 actually states that the server MAY continue to read
            # the request body after it has sent a final status code
            # (http://tools.ietf.org/html/rfc2616#section-8.2.3). However,
            # that totally defeats the purpose of 100-continue, so we hope
            # that the server behaves sanely and does not attempt to read
            # the body of a request it has already handled. (As a side note,
            # this ambiguity in the RFC also totally breaks HTTP pipelining,
            # as we can never be sure if the server is going to expect the
            # request or some request body data).
            self._out_remaining = None

        #
        # Prepare to read body
        #
        body_length = self._setup_read(method, status, header)

        # Don't require calls to co_read() et al if there is
        # nothing to be read.
        if self._in_remaining is None:
            self._pending_requests.popleft()

        log.debug('done (in_remaining=%s)', self._in_remaining)
        return HTTPResponse(method, path, status, reason, header, body_length)

    def _setup_read(self, method, status, header):
        '''Prepare for reading response body

        Sets up `._encoding`, `_in_remaining` and returns Content-Length
        (if available).

        See RFC 2616, sec. 4.4 for specific rules.
        '''

        # On error, the exception is stored in _encoding and raised on
        # the next call to co_read() et al - that way we can still
        # return the http status and headers.

        will_close = header.get('Connection', 'keep-alive').lower() == 'close'

        body_length = header['Content-Length']
        if body_length is not None:
            try:
                body_length = int(body_length)
            except ValueError:
                self._encoding = InvalidResponse('Invalid content-length: %s' % body_length)
                self._in_remaining = RESPONSE_BODY_ERROR
                return None

        if (
            status == NO_CONTENT
            or status == NOT_MODIFIED
            or 100 <= status < 200
            or method == 'HEAD'
        ):
            log.debug('no content by RFC')
            self._in_remaining = None
            self._encoding = None
            return 0

        tc = header.get('Transfer-Encoding', 'identity').lower()
        if tc == 'chunked':
            log.debug('Chunked encoding detected')
            self._encoding = Encodings.CHUNKED
            self._in_remaining = 0
            return None

        elif tc != 'identity':
            log.warning('Server uses invalid response encoding "%s"', tc)
            self._encoding = InvalidResponse('Cannot handle %s encoding' % tc)
            self._in_remaining = RESPONSE_BODY_ERROR
            return None

        log.debug('identity encoding detected')
        self._encoding = Encodings.IDENTITY

        if body_length is not None:
            log.debug('Will read response body of %d bytes', body_length)
            self._in_remaining = body_length or None
            return body_length

        if will_close:
            log.debug('no content-length, will read until EOF')
        else:
            log.warning(
                '%s:%d sent response with missing content-length header. This is not HTTP/1.1 '
                'specification compliant but we will ignore this error. Response headers: %s',
                self.hostname,
                self.port,
                header,
                extra={'rate_limit': 60},
            )
            # correct the connection header for library users (HTTPConnection does not need it)
            del header['Connection']
            header['Connection'] = 'close'
        self._in_remaining = READ_UNTIL_EOF
        return None

    async def _co_read_status(self):
        '''Read response line'''

        log.debug('start')

        # read status
        try:
            line = await self._co_readstr_until(b'\r\n', MAX_LINE_SIZE)
        except _ChunkTooLong:
            raise InvalidResponse('server send ridiculously long status line')

        try:
            version, status, reason = line.split(None, 2)
        except ValueError:
            try:
                version, status = line.split(None, 1)
                reason = ""
            except ValueError:
                # empty version will cause next test to fail.
                version = ""

        if not version.startswith("HTTP/1"):
            raise UnsupportedResponse('%s not supported' % version)

        # The status code is a three-digit number
        try:
            status = int(status)
            if status < 100 or status > 999:
                raise InvalidResponse('%d is not a valid status' % status)
        except ValueError:
            raise InvalidResponse('%s is not a valid status' % status)

        log.debug('done')
        return (status, reason.strip())

    async def _co_read_header(self):
        '''Read response header'''

        log.debug('start')

        # Peek into buffer. If the first characters are \r\n, then the header
        # is empty (so our search for \r\n\r\n would fail)
        rbuf = self._rbuf
        if len(rbuf) < 2:
            await self._co_fill_buffer(2)
        if rbuf.d[rbuf.b : rbuf.b + 2] == b'\r\n':
            log.debug('done (empty header)')
            rbuf.b += 2
            return ''

        try:
            hstring = await self._co_readstr_until(b'\r\n\r\n', MAX_HEADER_SIZE)
        except _ChunkTooLong:
            raise InvalidResponse('server sent ridiculously long header')

        log.debug('done (%d characters)', len(hstring))
        return hstring

    def read(self, len_=None):
        '''placeholder, will be replaced dynamically'''
        if len_ is None:
            return self.readall()
        orig_context = self._sync_context
        self._sync_context = True
        try:
            return eval_coroutine(self.co_read(len_))
        finally:
            self._sync_context = orig_context

    async def co_read(self, len_=None):
        '''Read up to *len_* bytes of response body data

        This method may return less than *len_* bytes, but will return ``b''`` only
        if the response body has been read completely.

        If *len_* is `None`, this method returns the entire response body.
        '''

        log.debug('start (len=%d)', len_)

        if self._in_remaining is RESPONSE_BODY_ERROR:
            raise self._encoding
        elif len_ is None:
            return await self.co_readall()
        elif len_ == 0 or self._in_remaining is None:
            return b''
        elif self._encoding is Encodings.IDENTITY:
            return await self._co_read_id(len_)
        elif self._encoding is Encodings.CHUNKED:
            return await self._co_read_chunked(len_=len_)
        else:
            raise RuntimeError('ooops, this should not be possible')

    def read_raw(self, size):
        '''Read *size* bytes of uninterpreted data

        This method may be used even after `UnsupportedResponse` or
        `InvalidResponse` has been raised. It reads raw data from the socket
        without attempting to interpret it. This is probably only useful for
        debugging purposes to take a look at the raw data received from the
        server. This method blocks if no data is available, and returns ``b''``
        if the connection has been closed.

        Calling this method will break the internal state and switch the socket
        to blocking operation. The connection has to be closed and reestablished
        afterwards.

        **Don't use this method unless you know exactly what you are doing**.
        '''

        if self._sock is None:
            raise ConnectionClosed('connection has been closed locally')

        self._sock.setblocking(True)

        buf = bytearray()
        rbuf = self._rbuf
        while len(buf) < size:
            len_ = min(size - len(buf), len(rbuf))
            if len_ < len(rbuf):
                buf += rbuf.d[rbuf.b : rbuf.b + len_]
                rbuf.b += len_
            elif len_ == 0:
                buf2 = self._sock.recv(size - len(buf))
                if not buf2:
                    break
                if self.trace_fh:
                    self.trace_fh.write(buf2)
                buf += buf2
            else:
                buf += rbuf.exhaust()

        return buf

    async def _co_read_id(self, len_):
        '''Read up to *len* bytes of response body assuming identity encoding'''

        log.debug('start (len=%d)', len_)
        assert self._in_remaining is not None

        if not self._in_remaining:
            # Body retrieved completely, clean up
            self._in_remaining = None
            self._pending_requests.popleft()
            return b''

        rbuf = self._rbuf
        if self._in_remaining is not READ_UNTIL_EOF:
            len_ = min(len_, self._in_remaining)
        log.debug('updated len_=%d', len_)

        # If buffer is empty, reset so that we start filling from
        # beginning. This check is already done by _try_fill_buffer(), but we
        # have to do it here or we never enter the while loop if the buffer
        # is empty but has no capacity (rbuf.b == rbuf.e == len(rbuf.d))
        if rbuf.b == rbuf.e:
            rbuf.b = 0
            rbuf.e = 0

        # Loop while we could return more data than we have buffered
        # and buffer is not full
        while len(rbuf) < len_ and rbuf.e < len(rbuf.d):
            got_data = self._try_fill_buffer()
            if got_data is None:
                if rbuf:
                    log.debug('nothing more to read')
                    break
                else:
                    log.debug('buffer empty and nothing to read, yielding..')
                if self._sync_context:
                    if not self._poll_recv(self._timeout_ms):
                        raise ConnectionTimedOut('Timeout in recv()')
                else:
                    with trio.move_on_after(self._timeout_ms / 1000) as ctx:
                        await trio.lowlevel.wait_readable(self._sock)
                    if ctx.cancelled_caught:
                        raise ConnectionTimedOut('Timeout in recv()')
            elif got_data == 0:
                if self._in_remaining is READ_UNTIL_EOF:
                    log.debug('connection closed, %d bytes in buffer', len(rbuf))
                    self._in_remaining = len(rbuf)
                    break
                else:
                    raise ConnectionClosed('server closed connection')

        len_ = min(len_, len(rbuf))
        if self._in_remaining is not READ_UNTIL_EOF:
            self._in_remaining -= len_

        if len_ < len(rbuf):
            buf = rbuf.d[rbuf.b : rbuf.b + len_]
            rbuf.b += len_
        else:
            buf = rbuf.exhaust()

        # When reading until EOF, it is possible that we read only the EOF. In
        # this case we won't get called again, so we need to clean-up the
        # request. This can not happen when we know the number of remaining
        # bytes, because in this case either the check at the start of the
        # function hits, or we can't read the remaining data and raise.
        if len(buf) == 0:
            assert self._in_remaining == 0
            self._in_remaining = None
            self._pending_requests.popleft()

        log.debug('done (%d bytes)', len(buf))
        return buf

    async def _co_read_chunked(self, len_: int):
        '''Read response body assuming chunked encoding'''

        assert isinstance(self._in_remaining, int)

        if self._in_remaining == 0:
            log.debug('starting next chunk')
            try:
                line = await self._co_readstr_until(b'\r\n', MAX_LINE_SIZE)
            except _ChunkTooLong:
                raise InvalidResponse('could not find next chunk marker')

            i = line.find(";")
            if i >= 0:
                log.debug('stripping chunk extensions: %s', line[i:])
                line = line[:i]  # strip chunk-extensions
            try:
                self._in_remaining = int(line, 16)
            except ValueError:
                raise InvalidResponse('Cannot read chunk size %r' % line[:20])

            log.debug('chunk size is %d', self._in_remaining)
            if self._in_remaining == 0:
                self._in_remaining = None
                self._pending_requests.popleft()

        if self._in_remaining is None:
            res = b''
        else:
            res = await self._co_read_id(len_)

        if not self._in_remaining:
            log.debug('chunk complete')
            await self._co_read_header()

        log.debug('done')
        return res

    async def _co_readstr_until(self, substr, maxsize):
        '''Read from server until *substr*, and decode to latin1

        If *substr* cannot be found in the next *maxsize* bytes,
        raises `_ChunkTooLong`.
        '''

        if not isinstance(substr, (bytes, bytearray, memoryview)):
            raise TypeError('*substr* must be bytes-like')

        log.debug('reading until %s', substr)

        rbuf = self._rbuf
        sub_len = len(substr)

        # Make sure that substr cannot be split over more than one part
        assert len(rbuf.d) > sub_len

        parts = []
        while True:
            # substr may be split between last part and current buffer
            # This isn't very performant, but it should be pretty rare
            if parts and sub_len > 1:
                buf = b''.join(
                    (parts[-1][-sub_len:], rbuf.d[rbuf.b : min(rbuf.e, rbuf.b + sub_len - 1)])
                )
                idx = buf.find(substr)
                if idx >= 0:
                    idx -= sub_len
                    break

            # log.debug('rbuf is: %s', rbuf.d[rbuf.b:min(rbuf.e, rbuf.b+512)])
            stop = min(rbuf.e, rbuf.b + maxsize)
            idx = rbuf.d.find(substr, rbuf.b, stop)

            if idx >= 0:  # found
                break
            if stop != rbuf.e:
                raise _ChunkTooLong()

            # If buffer is full, store away the part that we need for sure
            if rbuf.e == len(rbuf.d):
                log.debug('buffer is full, storing part')
                buf = rbuf.exhaust()
                parts.append(buf)
                maxsize -= len(buf)

            # Refill buffer
            while True:
                res = self._try_fill_buffer()
                if res is None:
                    log.debug('need more data, yielding')
                    if self._sync_context:
                        if not self._poll_recv(self._timeout_ms):
                            raise ConnectionTimedOut('Timeout in recv()')
                    else:
                        with trio.move_on_after(self._timeout_ms / 1000) as ctx:
                            await trio.lowlevel.wait_readable(self._sock)
                        if ctx.cancelled_caught:
                            raise ConnectionTimedOut('Timeout in recv()')
                elif res == 0:
                    raise ConnectionClosed('server closed connection')
                else:
                    break

        log.debug('found substr at %d', idx)
        idx += len(substr)
        buf = rbuf.d[rbuf.b : idx]
        rbuf.b = idx

        if parts:
            parts.append(buf)
            buf = b''.join(parts)

        try:
            return buf.decode('latin1')
        except UnicodeDecodeError:
            raise InvalidResponse('server response cannot be decoded to latin1')

    def _try_fill_buffer(self):
        '''Try to fill up read buffer

        Returns the number of bytes read into buffer, or `None` if no data was
        available on the socket. Return zero if the TCP connection has been
        properly closed but the socket object still exists. On other problems
        (e.g. if the socket object has been destroyed or the connection
        interrupted), raise `ConnectionClosed`.
        '''

        log.debug('start')
        rbuf = self._rbuf

        # If buffer is empty, reset so that we start filling from beginning
        if rbuf.b == rbuf.e:
            rbuf.b = 0
            rbuf.e = 0

        # There should be free capacity
        assert rbuf.e < len(rbuf.d)

        if self._sock is None:
            raise ConnectionClosed('connection has been closed locally')

        try:
            len_ = self._sock.recv_into(memoryview(rbuf.d)[rbuf.e :])
            if self.trace_fh:
                self.trace_fh.write(rbuf.d[rbuf.e : rbuf.e + len_])
        except (socket.timeout, ssl.SSLWantReadError, BlockingIOError):
            log.debug('done (nothing ready)')
            return None
        except (ConnectionResetError, BrokenPipeError, ssl.SSLEOFError):
            raise ConnectionClosed('connection was interrupted')

        rbuf.e += len_
        log.debug('done (got %d bytes)', len_)
        return len_

    async def _co_fill_buffer(self, len_):
        '''Make sure that there are at least *len_* bytes in buffer'''

        rbuf = self._rbuf
        if len_ > len(rbuf.d):
            raise ValueError('Requested more bytes than buffer has capacity')
        while len(rbuf) < len_:
            if len(rbuf.d) - rbuf.b < len_:
                self._rbuf.compact()
            res = self._try_fill_buffer()
            if res is None:
                if self._sync_context:
                    if not self._poll_recv(self._timeout_ms):
                        raise ConnectionTimedOut('Timeout in recv()')
                else:
                    with trio.move_on_after(self._timeout_ms / 1000) as ctx:
                        await trio.lowlevel.wait_readable(self._sock)
                    if ctx.cancelled_caught:
                        raise ConnectionTimedOut('Timeout in recv()')
            elif res == 0:
                raise ConnectionClosed('server closed connection')

    def readall(self):
        '''placeholder, will be replaced dynamically'''
        orig_context = self._sync_context
        self._sync_context = True
        try:
            return eval_coroutine(self.co_readall())
        finally:
            self._sync_context = orig_context

    async def co_readall(self):
        '''Read and return complete response body'''

        if self._in_remaining is None:
            return b''

        log.debug('start')
        parts = []
        while True:
            buf = await self.co_read(BUFFER_SIZE)
            log.debug('got %d bytes', len(buf))
            if not buf:
                break
            parts.append(buf)
        buf = b''.join(parts)
        log.debug('done (%d bytes)', len(buf))
        if self.trace_fh:
            self.trace_fh.write(buf)
        return buf

    def discard(self):
        '''placeholder, will be replaced dynamically'''
        orig_context = self._sync_context
        self._sync_context = True
        try:
            return eval_coroutine(self.co_discard())
        finally:
            self._sync_context = orig_context

    async def co_discard(self):
        '''Read and discard current response body'''

        if self._in_remaining is None:
            return

        log.debug('start')
        while True:
            buf = await self.co_read(BUFFER_SIZE)
            if not buf:
                break
            log.debug('discarding %d bytes', len(buf))
        log.debug('done')

    def reset(self):
        '''Reset HTTP connection

        This method resets the status of the HTTP connection. Any cached data and pending responses
        are discarded.
        '''
        self.disconnect()
        self.connect()

    def disconnect(self):
        '''Close HTTP connection'''

        log.debug('start')
        if self.trace_fh:
            self.trace_fh.close()
        if self._sock:
            self._sock.close()
            self._sock = None
            self._rbuf.clear()
            self._poll_recv = None
            self._poll_send = None
        else:
            log.debug('already closed')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.disconnect()
        return False


def _extend_HTTPConnection_docstrings():
    co_suffix = '\n\n' + textwrap.fill(
        'This method returns a coroutine. `.%s` is a regular method '
        'implementing the same functionality.',
        width=78,
    )
    reg_suffix = '\n\n' + textwrap.fill(
        'This method may block. `.co_%s` provides a coroutine '
        'implementing the same functionality without blocking.',
        width=78,
    )

    for name in (
        'read',
        'read_response',
        'readall',
        'send_request',
        'write',
        'discard',
        '_tunnel',
    ):
        fn = getattr(HTTPConnection, name)
        if name.startswith('_'):
            cofn = getattr(HTTPConnection, '_co' + name)
        else:
            cofn = getattr(HTTPConnection, 'co_' + name)

        fn.__doc__ = getdoc(cofn) + reg_suffix % name
        cofn.__doc__ = getdoc(cofn) + co_suffix % name


_extend_HTTPConnection_docstrings()


def eval_coroutine(coroutine):
    '''Retrieve the result of *coroutine*

    The coroutine must not yield.
    '''

    it = coroutine.__await__()
    try:
        next(it)
    except StopIteration as exc:
        return exc.value
    else:
        raise RuntimeError("coroutine suspended, can't retrieve result")


def create_socket(address):
    '''Create socket connected to *address**

    Use `socket.create_connection` to create a connected socket and return
    it. If a DNS related exception is raised, capture it and attempt to
    determine if the dns server is not reachable, or if the host name could
    not be resolved. Then raise either `NoSuchAddress` or
    `NameResolutionError` as appropriate.

    To distinguish between an unresolvable hostname and a problem with the
    DNS server, attempt to resolve the addresses in `DNS_TEST_HOSTNAMES`. If
    at least one test hostname can be resolved, assume that the DNS server
    is available and that *address* can not be resolved.
    '''

    def try_connect(fn, arg):
        try:
            return fn()
        except (socket.gaierror, socket.herror) as exc:
            if exc.errno in (socket.EAI_NODATA, socket.EAI_NONAME, socket.EAI_AGAIN):
                raise HostnameNotResolvableOrDNSUnavailable(arg)
            else:
                raise

    try:
        return try_connect(lambda: socket.create_connection(address), address[0])
    except HostnameNotResolvableOrDNSUnavailable:
        pass

    # Unfortunately, some exceptions do not help us to distinguish between
    # permanent and temporary problems (cf. https://stackoverflow.com/questions/24855168/,
    # https://stackoverflow.com/questions/24855669/). Therefore, we try to resolve
    # "well-known" hosts to resolve the ambiguity.

    for hostname, port in DNS_TEST_HOSTNAMES:
        try:
            try_connect(lambda: socket.getaddrinfo(hostname, port), address[0])
        except HostnameNotResolvableOrDNSUnavailable:
            pass
        else:
            # Reachable, now try to resolve original address again (maybe dns was only
            # down briefly)
            break
    else:
        # No host was reachable
        raise DNSUnavailable(address[0])

    # Try to connect to original host again
    try:
        return try_connect(lambda: socket.create_connection(address), address[0])
    except HostnameNotResolvableOrDNSUnavailable:
        raise HostnameNotResolvable(address[0])


def is_temp_network_error(exc):
    '''Return true if *exc* represents a potentially temporary network problem

    DNS resolution errors (`socket.gaierror` or `socket.herror`) are considered
    permanent, because `HTTPConnection` employs a heuristic to convert these
    exceptions to `HostnameNotResolvable` or `DNSUnavailable` instead.
    '''

    if isinstance(
        exc,
        (
            socket.timeout,
            ConnectionError,
            TimeoutError,
            InterruptedError,
            ConnectionClosed,
            ssl.SSLZeroReturnError,
            ssl.SSLEOFError,
            ssl.SSLSyscallError,
            ConnectionTimedOut,
            DNSUnavailable,
        ),
    ):
        return True

    elif isinstance(exc, OSError):
        # We have to be careful when retrieving errno codes, because
        # not all of them may exist on every platform.
        for errcode in (
            'EHOSTDOWN',
            'EHOSTUNREACH',
            'ENETDOWN',
            'ENETRESET',
            'ENETUNREACH',
            'ENOLINK',
            'ENONET',
            'ENOTCONN',
            'ENXIO',
            'EPIPE',
            'EREMCHG',
            'ESHUTDOWN',
            'ETIMEDOUT',
            'EAGAIN',
        ):
            try:
                if getattr(errno, errcode) == exc.errno:
                    return True
            except AttributeError:
                pass

    return False


class CaseInsensitiveDict(MutableMapping):
    """A case-insensitive `dict`-like object.

    Implements all methods and operations of
    :class:`collections.abc.MutableMapping` as well as `.copy`.

    All keys are expected to be strings. The structure remembers the case of the
    last key to be set, and :meth:`!iter`, :meth:`!keys` and :meth:`!items` will
    contain case-sensitive keys. However, querying and contains testing is
    case-insensitive::

        cid = CaseInsensitiveDict()
        cid['Accept'] = 'application/json'
        cid['aCCEPT'] == 'application/json' # True
        list(cid) == ['Accept'] # True

    For example, ``headers['content-encoding']`` will return the value of a
    ``'Content-Encoding'`` response header, regardless of how the header name
    was originally stored.

    If the constructor, :meth:`!update`, or equality comparison operations are
    given multiple keys that have equal lower-case representations, the behavior
    is undefined.
    """

    def __init__(self, data=None, **kwargs):
        self._store = dict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key, value):
        # Use the lowercased key for lookups, but store the actual
        # key alongside the value.
        self._store[key.lower()] = (key, value)

    def __getitem__(self, key):
        return self._store[key.lower()][1]

    def __delitem__(self, key):
        del self._store[key.lower()]

    def __iter__(self):
        return (casedkey for casedkey, _ in self._store.values())

    def __len__(self):
        return len(self._store)

    def lower_items(self):
        """Like :meth:`!items`, but with all lowercase keys."""
        return ((lowerkey, keyval[1]) for (lowerkey, keyval) in self._store.items())

    def __eq__(self, other):
        if isinstance(other, Mapping):
            other = CaseInsensitiveDict(other)
        else:
            return NotImplemented
        # Compare insensitively
        return dict(self.lower_items()) == dict(other.lower_items())

    # Copy is required
    def copy(self):
        return CaseInsensitiveDict(self._store.values())

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, dict(self.items()))



#!/usr/bin/env python3
'''
t4_adm.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

if __name__ == '__main__':
    import sys

    import pytest

    sys.exit(pytest.main([__file__] + sys.argv[1:]))

import os.path
import re
import shutil
import subprocess
import tempfile
import unittest
from argparse import Namespace

import pytest
from t4_fuse import TestFuse

from s3ql.backends import local
from s3ql.backends.common import CorruptedObjectError
from s3ql.backends.comprenc import ComprencBackend


@pytest.mark.usefixtures('pass_s3ql_cmd_argv', 'pass_reg_output')
class AdmTests(unittest.TestCase):
    def setUp(self):
        self.cache_dir = tempfile.mkdtemp(prefix='s3ql-cache-')
        self.backend_dir = tempfile.mkdtemp(prefix='s3ql-backend-')

        self.storage_url = 'local://' + self.backend_dir
        self.passphrase = 'oeut3d'

    def tearDown(self):
        shutil.rmtree(self.cache_dir)
        shutil.rmtree(self.backend_dir)

    def mkfs(self):
        proc = subprocess.Popen(
            self.s3ql_cmd_argv('mkfs.s3ql')
            + [
                '-L',
                'test fs',
                '--data-block-size',
                '500',
                '--authfile',
                '/dev/null',
                '--cachedir',
                self.cache_dir,
                '--quiet',
                self.storage_url,
            ],
            stdin=subprocess.PIPE,
            universal_newlines=True,
            stdout=subprocess.PIPE,
        )

        print(self.passphrase, file=proc.stdin)
        print(self.passphrase, file=proc.stdin)
        proc.stdin.close()
        stdout = proc.stdout.read()
        proc.stdout.close()
        self.assertEqual(proc.wait(), 0)

        return stdout

    def test_passphrase(self):
        self.mkfs()

        passphrase_new = 'sd982jhd'

        proc = subprocess.Popen(
            self.s3ql_cmd_argv('s3qladm')
            + [
                '--quiet',
                '--log',
                'none',
                '--authfile',
                '/dev/null',
                'passphrase',
                self.storage_url,
            ],
            stdin=subprocess.PIPE,
            universal_newlines=True,
        )

        print(self.passphrase, file=proc.stdin)
        print(passphrase_new, file=proc.stdin)
        print(passphrase_new, file=proc.stdin)
        proc.stdin.close()

        self.assertEqual(proc.wait(), 0)

        plain_backend = local.Backend(Namespace(storage_url=self.storage_url))
        backend = ComprencBackend(passphrase_new.encode(), ('zlib', 6), plain_backend)

        backend.fetch('s3ql_passphrase')  # will fail with wrong pw

    def test_clear(self):
        self.mkfs()

        proc = subprocess.Popen(
            self.s3ql_cmd_argv('s3qladm')
            + ['--quiet', '--log', 'none', '--authfile', '/dev/null', 'clear', self.storage_url],
            stdin=subprocess.PIPE,
            universal_newlines=True,
        )
        print('yes', file=proc.stdin)
        proc.stdin.close()
        self.assertEqual(proc.wait(), 0)

        plain_backend = local.Backend(Namespace(storage_url=self.storage_url))
        assert list(plain_backend.list()) == []

    def test_key_recovery(self):
        mkfs_output = self.mkfs()

        hit = re.search(
            r'^---BEGIN MASTER KEY---\n' r'(.+)\n' r'---END MASTER KEY---$',
            mkfs_output,
            re.MULTILINE,
        )
        assert hit
        master_key = hit.group(1)

        plain_backend = local.Backend(Namespace(storage_url=self.storage_url))
        del plain_backend['s3ql_passphrase']  # Oops

        backend = ComprencBackend(self.passphrase.encode(), ('zlib', 6), plain_backend)
        with pytest.raises(CorruptedObjectError):
            backend.fetch('s3ql_params')

        passphrase_new = 'sd982jhd'

        proc = subprocess.Popen(
            self.s3ql_cmd_argv('s3qladm')
            + [
                '--quiet',
                '--log',
                'none',
                '--authfile',
                '/dev/null',
                'recover-key',
                self.storage_url,
            ],
            stdin=subprocess.PIPE,
            universal_newlines=True,
        )

        print(master_key, file=proc.stdin)
        print(passphrase_new, file=proc.stdin)
        print(passphrase_new, file=proc.stdin)
        proc.stdin.close()
        self.assertEqual(proc.wait(), 0)

        backend = ComprencBackend(passphrase_new.encode(), ('zlib', 6), plain_backend)

        backend.fetch('s3ql_passphrase')  # will fail with wrong pw


class TestMetadataRestore(TestFuse):
    def restore_backup(self):
        proc = subprocess.Popen(
            self.s3ql_cmd_argv('s3qladm')
            + [
                '--quiet',
                '--log',
                'none',
                '--authfile',
                '/dev/null',
                '--cachedir',
                self.cache_dir,
                'restore-metadata',
                self.storage_url,
            ],
            stdin=subprocess.PIPE,
            universal_newlines=True,
        )
        if self.backend_login is not None:
            print(self.backend_login, file=proc.stdin)
            print(self.backend_passphrase, file=proc.stdin)
        if self.passphrase is not None:
            print(self.passphrase, file=proc.stdin)
        # Restore second metadata backup
        print('1', file=proc.stdin)
        proc.stdin.close()
        assert proc.wait() == 0

    def test(self):
        self.mkfs()
        self.mount()
        with open(self.mnt_dir + "/mount1.txt", 'w') as fh:
            print('Data written on first mount', file=fh)
        self.umount()

        self.mount()
        with open(self.mnt_dir + "/mount2.txt", 'w') as fh:
            print('Data written on second mount', file=fh)
        self.umount()

        self.restore_backup()

        self.reg_output(
            r'^WARNING: Deleted spurious object \d+$',
            count=1,
        )
        self.fsck(expect_retcode=128)

        self.mount()
        assert not os.path.exists(self.mnt_dir + '/mount2.txt')
        assert os.path.exists(self.mnt_dir + '/mount1.txt')



'''
backends/__init__.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

from . import gs, local, rackspace, s3, s3c, s3c4, swift, swiftks
from .b2.b2_backend import B2Backend

#: Mapping from storage URL prefixes to backend classes
prefix_map = {
    's3': s3.Backend,
    'local': local.Backend,
    'gs': gs.Backend,
    's3c': s3c.Backend,
    's3c4': s3c4.Backend,
    'swift': swift.Backend,
    'swiftks': swiftks.Backend,
    'rackspace': rackspace.Backend,
    'b2': B2Backend,
}

__all__ = ['common', 'pool', 'comprenc'] + list(prefix_map.keys())



#!/usr/bin/env python3
'''
This file reproduces mount.s3ql, with pyfuse3.invalidate_inode patched
for the t5_cp.py::TestCp::test_cp_inode_invalidate test.
'''

import os.path
import sys

basedir = os.path.abspath(os.path.join(os.path.dirname(sys.argv[0]), '..'))
sys.path = [os.path.join(basedir, 'src')] + sys.path

# Override pyfuse3.invalidate_inode : Drop kernel dentries and inodes cache
# just before calling pyfuse3.invalidate_inode
import pyfuse3

pyfuse3_invalidate_inode = pyfuse3.invalidate_inode


def patched_pyfuse3_invalidate_inode(inode):
    # echo 2 > /proc/sys/vm/drop_caches : Drop kernel dentries and inodes cache
    with open("/proc/sys/vm/drop_caches", "w") as drop_caches:
        drop_caches.write("2\n")
    pyfuse3_invalidate_inode(inode)


pyfuse3.invalidate_inode = patched_pyfuse3_invalidate_inode

import s3ql.mount

s3ql.mount.main(sys.argv[1:])



'''
oauth_client.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import logging
import sys
import textwrap

from google_auth_oauthlib.flow import InstalledAppFlow

from .common import OAUTH_CLIENT_ID, OAUTH_CLIENT_SECRET
from .logging import setup_logging, setup_warnings
from .parse_args import ArgumentParser

log = logging.getLogger(__name__)


def parse_args(args):
    '''Parse command line'''

    parser = ArgumentParser(
        description=textwrap.dedent(
            '''\
        Obtain OAuth2 refresh token for Google Storage
        '''
        )
    )

    parser.add_log()
    parser.add_debug()
    parser.add_quiet()
    parser.add_version()

    return parser.parse_args(args)


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    setup_warnings()
    options = parse_args(args)
    setup_logging(options)

    # We need full control in order to be able to update metadata
    # cf. https://stackoverflow.com/questions/24718787
    flow = InstalledAppFlow.from_client_config(
        client_config={
            "installed": {
                "client_id": OAUTH_CLIENT_ID,
                "client_secret": OAUTH_CLIENT_SECRET,
                "redirect_uris": ["http://localhost", "urn:ietf:wg:oauth:2.0:oob"],
                "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                "token_uri": "https://accounts.google.com/o/oauth2/token",
            }
        },
        scopes=['https://www.googleapis.com/auth/devstorage.full_control'],
    )

    credentials = flow.run_local_server(open_browser=True)

    print('Success. Your refresh token is:\n', credentials.refresh_token)



#!/usr/bin/env python3
'''
t1_backends.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

if __name__ == '__main__':
    import sys

    import pytest

    sys.exit(pytest.main([__file__] + sys.argv[1:]))

import functools
import logging
import re
import shutil
import tempfile
import threading
import time
from argparse import Namespace
from io import BytesIO

import mock_server
import pytest
from common import CLOCK_GRANULARITY, NoTestSection, get_remote_test_info
from pytest import raises as assert_raises
from pytest_checklogs import assert_logs

from s3ql import BUFSIZE, backends
from s3ql.backends.common import AbstractBackend, CorruptedObjectError, NoSuchObject
from s3ql.backends.comprenc import ComprencBackend, ObjectNotEncrypted
from s3ql.backends.gs import Backend as GSBackend
from s3ql.backends.local import Backend as LocalBackend
from s3ql.backends.s3c import BadDigestError, HTTPError, OperationAbortedError, S3Error
from s3ql.http import ConnectionClosed

log = logging.getLogger(__name__)
empty_set = set()


def brace_expand(s):
    hit = re.search(r'^(.*)\{(.+)\}(.*)$', s)
    if not hit:
        return [s]
    (p, e, s) = hit.groups()
    l = []
    for el in e.split(','):
        l.append(p + el + s)
    return l


def enable_temp_fail(backend):
    if isinstance(backend, ComprencBackend):
        backend = backend.backend
    backend.unittest_info.may_temp_fail = True


# It'd be nice if we could use the setup_module hook instead, but
# unfortunately that gets executed *after* pytest_generate_tests.
def _get_backend_info():
    '''Get information about raw backends available for testing'''

    info = []

    # Local backend
    bi = Namespace()
    bi.name = 'local'
    bi.classname = 'local'
    info.append(bi)

    # Backends talking to actual remote servers (if available)
    for name in backends.prefix_map:
        if name == 'local':  # local backend has already been handled
            continue

        try:
            (login, password, storage_url) = get_remote_test_info(name + '-test')
        except NoTestSection as exc:
            log.info('Not running remote tests for %s backend: %s', name, exc.reason)
            continue

        bi = Namespace()
        bi.name = 'remote-' + name
        bi.classname = name
        bi.storage_url = storage_url
        bi.login = login
        bi.password = password
        info.append(bi)

    # Backends talking to local mock servers
    for request_handler, storage_url in mock_server.handler_list:
        name = re.match(r'^([a-zA-Z0-9]+)://', storage_url).group(1)
        bi = Namespace()
        bi.name = 'mock-' + name
        bi.classname = name
        bi.request_handler = request_handler
        bi.storage_url = storage_url
        info.append(bi)

    return info


def pytest_generate_tests(metafunc, _info_cache=[]):
    if _info_cache:
        backend_info = _info_cache[0]
    else:
        backend_info = _get_backend_info()
        _info_cache.append(backend_info)

    if 'backend' not in metafunc.fixturenames:
        return

    with_backend_mark = metafunc.definition.get_closest_marker('with_backend')
    assert with_backend_mark

    test_params = []
    for spec in with_backend_mark.args:
        (backend_spec, comprenc_spec) = spec.split('/')

        # Expand compression/encryption specification
        # raw == don't use ComprencBackend at all
        # plain = use ComprencBackend without compression and encryption
        if comprenc_spec == '*':
            comprenc_kinds = ['aes', 'aes+zlib', 'plain', 'zlib', 'bzip2', 'lzma', 'raw']
        else:
            comprenc_kinds = brace_expand(comprenc_spec)

        # Expand backend specification
        if backend_spec == '*':
            test_bi = backend_info
        else:
            test_bi = []
            for name in brace_expand(backend_spec):
                test_bi += [x for x in backend_info if x.classname == name]

        # Filter
        if with_backend_mark.kwargs.get('require_mock_server', False):
            test_bi = [x for x in test_bi if 'request_handler' in x]

        for comprenc_kind in comprenc_kinds:
            for bi in test_bi:
                test_params.append((bi, comprenc_kind))

    metafunc.parametrize(
        "backend",
        test_params,
        indirect=True,
        ids=['%s/%s' % (x[0].name, x[1]) for x in test_params],
    )


@pytest.fixture()
def backend(request):
    (backend_info, comprenc_kind) = request.param

    if backend_info.classname == 'local':
        gen = yield_local_backend(backend_info)
    elif 'request_handler' in backend_info:
        gen = yield_mock_backend(backend_info)
    else:
        gen = yield_remote_backend(backend_info)

    for raw_backend in gen:
        if comprenc_kind == 'raw':
            backend = raw_backend
        elif comprenc_kind == 'plain':
            backend = ComprencBackend(None, (None, 6), raw_backend)
        elif comprenc_kind == 'aes+zlib':
            backend = ComprencBackend(b'schlurz', ('zlib', 6), raw_backend)
        elif comprenc_kind == 'aes':
            backend = ComprencBackend(b'schlurz', (None, 6), raw_backend)
        else:
            backend = ComprencBackend(None, (comprenc_kind, 6), raw_backend)

        backend.unittest_info = raw_backend.unittest_info

        yield backend


def yield_local_backend(bi):
    backend_dir = tempfile.mkdtemp(prefix='s3ql-backend-')
    backend = LocalBackend(Namespace(storage_url='local://' + backend_dir))
    backend.unittest_info = Namespace()
    try:
        yield backend
    finally:
        backend.close()
        shutil.rmtree(backend_dir)


def yield_mock_backend(bi):
    backend_class = backends.prefix_map[bi.classname]
    server = mock_server.StorageServer(bi.request_handler, ('localhost', 0))
    thread = threading.Thread(target=server.serve_forever)
    thread.daemon = True
    thread.start()

    storage_url = bi.storage_url % {
        'host': server.server_address[0],
        'port': server.server_address[1],
    }
    backend = backend_class(
        Namespace(
            storage_url=storage_url,
            backend_login='joe',
            backend_password='swordfish',
            backend_options={'no-ssl': True},
        )
    )

    # Enable OAuth when using Google Backend
    if isinstance(backend, GSBackend):
        backend.use_oauth2 = True
        backend.hdr_prefix = 'x-goog-'  # Normally set in __init__
        backend.access_token[backend.password] = 'foobar'

    backend.unittest_info = Namespace()

    # Mock server should not have temporary failures by default
    is_temp_failure = backend.is_temp_failure

    @functools.wraps(backend.is_temp_failure)
    def wrap(exc):
        return backend.unittest_info.may_temp_fail and is_temp_failure(exc)

    backend.is_temp_failure = wrap
    backend.unittest_info.may_temp_fail = False

    try:
        yield backend
    finally:
        backend.close()
        server.shutdown()
        server.server_close()


def yield_remote_backend(bi, _ctr=[0]):
    # Add timestamp + ctr to prefix so we don't have to deal with cruft from
    # previous tests
    _ctr[0] += 1
    storage_url = bi.storage_url
    if storage_url[-1] != '/':
        storage_url += '/'
    storage_url += '%d_%d/' % (time.time(), _ctr[0])

    backend_class = backends.prefix_map[bi.classname]
    backend = backend_class(
        Namespace(
            storage_url=storage_url,
            backend_login=bi.login,
            backend_password=bi.password,
            backend_options={},
        )
    )

    backend.unittest_info = Namespace()
    try:
        yield backend
    finally:
        for name in list(backend.list()):
            backend.delete(name)
        backend.close()


def newname(name_counter=[0]):
    '''Return random, unique string'''
    name_counter[0] += 1
    return "random\\'name' %d" % name_counter[0]


def newvalue():
    return newname().encode()


def assert_in_index(backend, keys):
    '''Assert that *keys* are in backend list output'''

    keys = set(keys)
    index = set(backend.list())
    assert keys - index == empty_set


def assert_not_in_index(backend, keys):
    '''Assert that *keys* are not in backend list output'''

    keys = set(keys)
    index = set(backend.list())
    assert keys - index == keys


@pytest.mark.with_backend('*/raw', 'local/*')
def test_readinto_write_fh(backend: AbstractBackend):
    key = newname()
    value = newvalue()
    metadata = {'jimmy': 'jups@42'}
    buf = BytesIO()

    assert key not in backend
    assert_raises(NoSuchObject, backend.lookup, key)
    assert_raises(NoSuchObject, backend.readinto_fh, key, buf)

    assert backend.write_fh(key, BytesIO(value), metadata) > 0

    assert key in backend

    metadata2 = backend.readinto_fh(key, buf)

    assert value == buf.getvalue()
    assert metadata == metadata2
    assert backend.lookup(key) == metadata


@pytest.mark.with_backend('*/{plain,raw,aes+zlib}')
def test_write_fh_partial(backend: AbstractBackend):
    key = newname()
    metadata = {'jimmy': 'jups@42'}
    data = (''.join(str(x) for x in range(100))).encode()
    buf = BytesIO(data)

    buf.seek(10)
    assert backend.write_fh(key, buf, metadata, len_=20) > 0

    buf2 = BytesIO()
    backend.readinto_fh(key, buf2)

    assert buf2.getvalue() == data[10 : 10 + 20]


@pytest.mark.with_backend('*/{plain,raw,aes+zlib}')
def test_write_fh_off(backend: AbstractBackend):
    key = newname()
    metadata = {'jimmy': 'jups@42'}
    data = (''.join(str(x) for x in range(100))).encode()
    buf = BytesIO(data)

    buf.seek(10)
    assert backend.write_fh(key, buf, metadata) > 0

    buf2 = BytesIO()
    backend.readinto_fh(key, buf2)

    assert buf2.getvalue() == data[10:]


@pytest.mark.with_backend('swift/raw')
def test_issue114(backend, monkeypatch):
    key = newname()
    value = newvalue()
    monkeypatch.setitem(backend.options, 'disable-expect100', True)
    backend[key] = value
    assert_in_index(backend, [key])


@pytest.mark.with_backend('*/raw', 'local/aes+zlib')
def test_complex_meta(backend):
    key = newname()
    value = newvalue()

    metadata = {
        'com\nplex: key': 42,
        'farp_': False,
        'non-farp': True,
        'blu%rz': 23.283475,
        'görp': b'heelo',
        'sch.al': 'gorroobalp\nfurrö!',
        'lo-ng': 'foobarz' * 40,
    }

    assert key not in backend
    backend.store(key, value, metadata)
    (value2, metadata2) = backend.fetch(key)

    assert value == value2
    assert metadata == metadata2
    assert backend.lookup(key) == metadata


# No need to run with different encryption/compression settings,
# ComprencBackend should just forward this 1:1 to the raw backend.
@pytest.mark.with_backend('*/aes')
def test_list(backend):
    keys = ['prefixa' + newname() for _ in range(6)] + ['prefixb' + newname() for _ in range(6)]
    values = [newvalue() for _ in range(12)]

    assert set(backend.list()) == empty_set
    for i in range(12):
        backend[keys[i]] = values[i]
    assert_in_index(backend, keys)

    assert set(backend.list('prefixa')) == set(keys[:6])
    assert set(backend.list('prefixb')) == set(keys[6:])
    assert set(backend.list('prefixc')) == empty_set


# No need to run with different encryption/compression settings,
# ComprencBackend should just forward this 1:1 to the raw backend.
@pytest.mark.with_backend('*/aes')
def test_delete(backend):
    key = newname()
    value = newvalue()

    backend[key] = value
    assert_in_index(backend, [key])
    backend.fetch(key)

    del backend[key]

    assert_not_in_index(backend, [key])
    with pytest.raises(NoSuchObject):
        backend.fetch(key)


@pytest.mark.with_backend('*/aes')
def test_delete_non_existing(backend):
    key = newname()
    del backend[key]


@pytest.mark.with_backend('b2/aes')
def test_delete_b2_versions(backend):
    key = newname()
    values = [newvalue() for _ in range(3)]

    # Write values to same key, should become versions of the key
    for value in values:
        backend[key] = value

    # Wait for it
    assert_in_index(backend, [key])
    backend.fetch(key)

    # Delete it, should delete all versions
    del backend[key]

    assert_not_in_index(backend, [key])
    with pytest.raises(NoSuchObject):
        backend.fetch(key)


# No need to run with different encryption/compression settings,
# ComprencBackend should just forward this 1:1 to the raw backend.
@pytest.mark.with_backend('*/aes')
def test_delete_multi(backend):
    if not backend.has_delete_multi:
        pytest.skip('backend does not support delete_multi')

    keys = [newname() for _ in range(10)]
    value = newvalue()

    for key in keys:
        backend[key] = value

    # Delete half of them
    to_delete = keys[::2]
    backend.delete_multi(to_delete)
    assert len(to_delete) == 0

    assert_not_in_index(backend, keys[::2])
    for key in keys[::2]:
        with pytest.raises(NoSuchObject):
            backend.fetch(key)

    assert_in_index(backend, keys[1::2])
    for key in keys[1::2]:
        backend.fetch(key)


@pytest.mark.with_backend('local/{aes,aes+zlib,zlib,bzip2,lzma}')
def test_corruption(backend):
    plain_backend = backend.backend

    # Create compressed object
    key = newname()
    value = newvalue()
    backend[key] = value

    # Retrieve compressed data
    (compr_value, meta) = plain_backend.fetch(key)
    compr_value = bytearray(compr_value)

    # Overwrite with corrupted data
    compr_value[-3:] = b'000'
    plain_backend.store(key, compr_value, meta)

    with pytest.raises(CorruptedObjectError) as exc:
        backend.fetch(key)

    if backend.passphrase is None:  # compression only
        assert exc.value.str == 'Invalid compressed stream'
    else:
        assert exc.value.str == 'HMAC mismatch'


@pytest.mark.with_backend('local/{aes,aes+zlib,zlib,bzip2,lzma}')
def test_extra_data(backend):
    plain_backend = backend.backend

    # Create compressed object
    key = newname()
    value = newvalue()
    backend[key] = value

    # Retrieve compressed data
    (compr_value, meta) = plain_backend.fetch(key)
    compr_value = bytearray(compr_value)

    # Overwrite with extended data
    compr_value += b'000'
    plain_backend.store(key, compr_value, meta)

    with pytest.raises(CorruptedObjectError) as exc:
        backend.fetch(key)

    if backend.passphrase is None:  # compression only
        assert exc.value.str == 'Data after end of compressed stream'
    else:
        assert exc.value.str == 'Extraneous data at end of object'


@pytest.mark.with_backend('local/{aes,aes+zlib}')
def test_encryption(backend):
    plain_backend = backend.backend

    plain_backend['plain'] = b'foobar452'
    backend.store('encrypted', b'testdata', {'tag': True})

    assert plain_backend['encrypted'] != b'testdata'
    assert_raises(CorruptedObjectError, backend.fetch, 'plain')
    assert_raises(CorruptedObjectError, backend.lookup, 'plain')

    backend.passphrase = None
    backend.store('not-encrypted', b'testdata2395', {'tag': False})
    assert_raises(CorruptedObjectError, backend.fetch, 'encrypted')
    assert_raises(CorruptedObjectError, backend.lookup, 'encrypted')

    backend.passphrase = b'jobzrul'
    assert_raises(CorruptedObjectError, backend.fetch, 'encrypted')
    assert_raises(CorruptedObjectError, backend.lookup, 'encrypted')
    assert_raises(ObjectNotEncrypted, backend.fetch, 'not-encrypted')
    assert_raises(ObjectNotEncrypted, backend.lookup, 'not-encrypted')


@pytest.mark.with_backend('local/{aes,aes+zlib}')
def test_replay(backend):
    plain_backend = backend.backend

    # Create encrypted object
    key1 = newname()
    key2 = newname()
    value = newvalue()
    backend[key1] = value

    # Retrieve compressed data
    (compr_value, meta) = plain_backend.fetch(key1)
    compr_value = bytearray(compr_value)

    # Copy into new object
    plain_backend.store(key2, compr_value, meta)

    with pytest.raises(CorruptedObjectError):
        backend.fetch(key2)


@pytest.mark.with_backend('s3c/raw', require_mock_server=True)
def test_list_bug(backend, monkeypatch):
    keys = ['prefixa' + newname() for _ in range(6)] + ['prefixb' + newname() for _ in range(6)]
    values = [newvalue() for _ in range(12)]

    assert set(backend.list()) == empty_set
    for i in range(12):
        backend[keys[i]] = values[i]
    assert_in_index(backend, keys)

    # Force reconnect during list
    handler_class = mock_server.S3CRequestHandler

    def do_list(self, q, _real=handler_class.do_list):
        q.params['max_keys'] = ['3']
        self.close_connection = True
        return _real(self, q)

    monkeypatch.setattr(handler_class, 'do_list', do_list)

    enable_temp_fail(backend)
    assert set(backend.list()) == set(keys)


@pytest.mark.with_backend('s3c/aes+zlib', require_mock_server=True)
def test_corrupted_get(backend, monkeypatch):
    key = 'brafasel'
    value = b'hello there, let us see whats going on'
    backend[key] = value

    # Monkeypatch request handler to produce invalid etag
    handler_class = mock_server.S3CRequestHandler

    def send_header(self, keyword, value, count=[0], send_header_real=handler_class.send_header):
        if keyword == 'ETag':
            count[0] += 1
            if count[0] <= 3:
                value = value[::-1]
        return send_header_real(self, keyword, value)

    monkeypatch.setattr(handler_class, 'send_header', send_header)

    with assert_logs('^MD5 mismatch for', count=1, level=logging.WARNING):
        assert_raises(BadDigestError, backend.fetch, key)

    enable_temp_fail(backend)
    with assert_logs('^MD5 mismatch for', count=2, level=logging.WARNING):
        assert backend[key] == value


@pytest.mark.with_backend('s3c/{plain,raw,aes+zlib}', require_mock_server=True)
def test_corrupted_meta(backend, monkeypatch):
    key = 'brafasel'
    value = b'hello there, let us see whats going on'
    backend[key] = value

    # Monkeypatch request handler to mess up metadata
    handler_class = mock_server.S3CRequestHandler

    def send_header(self, keyword, value, count=[0], send_header_real=handler_class.send_header):
        if keyword == self.hdr_prefix + 'Meta-md5':
            count[0] += 1
            if count[0] <= 3:
                value = value[::-1]
        return send_header_real(self, keyword, value)

    monkeypatch.setattr(handler_class, 'send_header', send_header)

    with assert_logs('^MD5 mismatch in metadata for', count=1, level=logging.WARNING):
        assert_raises(BadDigestError, backend.fetch, key)

    enable_temp_fail(backend)
    with assert_logs('^MD5 mismatch in metadata for', count=2, level=logging.WARNING):
        assert backend[key] == value


@pytest.mark.with_backend('s3c/{plain,raw,aes+zlib}', require_mock_server=True)
def test_corrupted_put(backend, monkeypatch):
    key = 'brafasel'
    value = b'hello there, let us see whats going on'

    # Monkeypatch request handler to produce invalid etag
    handler_class = mock_server.S3CRequestHandler

    def send_header(self, keyword, value, count=[0], send_header_real=handler_class.send_header):
        if keyword == 'ETag':
            count[0] += 1
            if count[0] < 3:
                value = value[::-1]
        return send_header_real(self, keyword, value)

    monkeypatch.setattr(handler_class, 'send_header', send_header)

    with pytest.raises(BadDigestError):
        backend.store(key, value)

    enable_temp_fail(backend)
    backend.store(key, value)

    assert backend[key] == value


@pytest.mark.with_backend('s3c/{plain,raw,aes+zlib}', require_mock_server=True)
def test_get_s3error(backend, monkeypatch):
    value = b'hello there, let us see whats going on'
    key = 'quote'
    backend[key] = value

    # Monkeypatch request handler to produce 3 errors
    handler_class = mock_server.S3CRequestHandler

    def do_GET(self, real_GET=handler_class.do_GET, count=[0]):
        count[0] += 1
        if count[0] > 3:
            return real_GET(self)
        else:
            self.send_error(503, code='OperationAborted')

    monkeypatch.setattr(handler_class, 'do_GET', do_GET)
    assert_raises(OperationAbortedError, backend.fetch, value)

    enable_temp_fail(backend)
    assert backend[key] == value


@pytest.mark.with_backend('s3c/{plain,raw,aes+zlib}', require_mock_server=True)
def test_head_s3error(backend, monkeypatch):
    value = b'hello there, let us see whats going on'
    key = 'quote'
    meta = {'bar': 42, 'foo': 42**2}
    backend.store(key, value, metadata=meta)

    # Monkeypatch request handler to produce 3 errors
    handler_class = mock_server.S3CRequestHandler

    def do_HEAD(self, real_HEAD=handler_class.do_HEAD, count=[0]):
        count[0] += 1
        if count[0] > 3:
            return real_HEAD(self)
        else:
            self.send_error(503, code='OperationAborted')

    monkeypatch.setattr(handler_class, 'do_HEAD', do_HEAD)
    with pytest.raises(HTTPError) as exc:
        backend.lookup(key)
    assert exc.value.status == 503

    enable_temp_fail(backend)
    assert backend.lookup(key) == meta


@pytest.mark.with_backend('s3c/raw', require_mock_server=True)
def test_delete_s3error(backend, monkeypatch):
    value = b'hello there, let us see whats going on'
    key = 'quote'
    backend[key] = value

    # Monkeypatch request handler to produce 3 errors
    handler_class = mock_server.S3CRequestHandler

    def do_DELETE(self, real_DELETE=handler_class.do_DELETE, count=[0]):
        count[0] += 1
        if count[0] > 3:
            return real_DELETE(self)
        else:
            self.send_error(503, code='OperationAborted')

    monkeypatch.setattr(handler_class, 'do_DELETE', do_DELETE)
    assert_raises(OperationAbortedError, backend.delete, key)

    enable_temp_fail(backend)
    backend.delete(key)


@pytest.mark.with_backend('s3c/raw', require_mock_server=True)
def test_backoff(backend, monkeypatch):
    value = b'hello there, let us see whats going on'
    key = 'quote'
    backend[key] = value

    # Monkeypatch request handler
    handler_class = mock_server.S3CRequestHandler
    timestamps = []

    def do_DELETE(self, real_DELETE=handler_class.do_DELETE):
        timestamps.append(time.time())
        if len(timestamps) < 3:
            self.send_error(503, code='SlowDown', extra_headers={'Retry-After': '1'})
        else:
            return real_DELETE(self)

    monkeypatch.setattr(handler_class, 'do_DELETE', do_DELETE)
    enable_temp_fail(backend)
    backend.delete(key)

    assert timestamps[1] - timestamps[0] > 1 - CLOCK_GRANULARITY
    assert timestamps[2] - timestamps[1] > 1 - CLOCK_GRANULARITY
    assert timestamps[2] - timestamps[0] < 10


@pytest.mark.with_backend('s3c/aes+zlib', require_mock_server=True)
def test_httperror(backend, monkeypatch):
    value = b'hello there, let us see whats going on'
    key = 'quote'
    backend[key] = value

    # Monkeypatch request handler to produce a HTTP Error
    handler_class = mock_server.S3CRequestHandler

    def do_DELETE(self, real_DELETE=handler_class.do_DELETE, count=[0]):
        count[0] += 1
        if count[0] >= 3:
            return real_DELETE(self)
        content = "I'm a proxy, and I messed up!".encode('utf-8')
        self.send_response(502, "Bad Gateway")
        self.send_header("Content-Type", 'text/plain; charset="utf-8"')
        self.send_header("Content-Length", str(len(content)))
        self.end_headers()
        if self.command != 'HEAD':
            self.wfile.write(content)

    monkeypatch.setattr(handler_class, 'do_DELETE', do_DELETE)
    assert_raises(HTTPError, backend.delete, key)

    enable_temp_fail(backend)
    backend.delete(key)


@pytest.mark.with_backend('s3c/aes+zlib', require_mock_server=True)
def test_put_s3error_early(backend, monkeypatch):
    '''Fail after expect-100'''

    data = b'hello there, let us see whats going on'
    key = 'borg'

    # Monkeypatch request handler to produce 3 errors
    handler_class = mock_server.S3CRequestHandler

    def handle_expect_100(self, real=handler_class.handle_expect_100, count=[0]):
        count[0] += 1
        if count[0] > 3:
            return real(self)
        else:
            self.send_error(503, code='OperationAborted')
            return False

    monkeypatch.setattr(handler_class, 'handle_expect_100', handle_expect_100)
    with pytest.raises(OperationAbortedError):
        backend.store(key, data)

    enable_temp_fail(backend)
    backend.store(key, data)


@pytest.mark.with_backend('s3c/aes+zlib', require_mock_server=True)
def test_put_s3error_med(backend, monkeypatch):
    '''Fail as soon as data is received'''
    data = b'hello there, let us see whats going on'
    key = 'borg'

    # Monkeypatch request handler to produce 3 errors
    handler_class = mock_server.S3CRequestHandler

    def do_PUT(self, real_PUT=handler_class.do_PUT, count=[0]):
        count[0] += 1
        # Note: every time we return an error, the request will be retried
        # *twice*: once because of the error, and a second time because the
        # connection has been closed by the server.
        if count[0] > 2:
            return real_PUT(self)
        else:
            self.send_error(503, code='OperationAborted')

            # Since we don't read all the data, we have to close
            # the connection
            self.close_connection = True

    monkeypatch.setattr(handler_class, 'do_PUT', do_PUT)
    with pytest.raises(OperationAbortedError):
        backend.store(key, data)

    enable_temp_fail(backend)
    backend.store(key, data)


@pytest.mark.with_backend('s3c/aes+zlib', require_mock_server=True)
def test_put_s3error_late(backend, monkeypatch):
    '''Fail after reading all data'''
    data = b'hello there, let us see whats going on'
    key = 'borg'

    # Monkeypatch request handler to produce 3 errors
    handler_class = mock_server.S3CRequestHandler

    def do_PUT(self, real_PUT=handler_class.do_PUT, count=[0]):
        count[0] += 1
        if count[0] > 3:
            return real_PUT(self)
        else:
            self.rfile.read(int(self.headers['Content-Length']))
            self.send_error(503, code='OperationAborted')

    monkeypatch.setattr(handler_class, 'do_PUT', do_PUT)
    with pytest.raises(OperationAbortedError):
        backend.store(key, data)

    enable_temp_fail(backend)
    backend.store(key, data)


@pytest.mark.with_backend('s3c/aes+zlib', require_mock_server=True)
def test_issue58(backend, monkeypatch):
    '''Send error while client is sending data'''

    # Monkeypatch request handler
    handler_class = mock_server.S3CRequestHandler

    def do_PUT(self, real=handler_class.do_PUT, count=[0]):
        count[0] += 1
        if count[0] > 1:
            return real(self)

        # Read half the data
        self.rfile.read(min(BUFSIZE, int(self.headers['Content-Length']) // 2))

        # Then generate an error and close the connection
        self.send_error(401, code='MalformedXML')
        self.close_connection = True

    monkeypatch.setattr(handler_class, 'do_PUT', do_PUT)

    # Write a big object. We need to write random data, or
    # compression while make the payload too small
    with open('/dev/urandom', 'rb') as rnd:
        buf = rnd.read(5 * BUFSIZE)
    with pytest.raises(S3Error) as exc_info:
        backend.store('borg', buf)
    assert exc_info.value.code == 'MalformedXML'

    enable_temp_fail(backend)
    backend.store('borg', buf)


@pytest.mark.with_backend('s3c/aes+zlib', require_mock_server=True)
def test_issue58_b(backend, monkeypatch):
    '''Close connection while client is sending data'''

    # Monkeypatch request handler
    handler_class = mock_server.S3CRequestHandler

    def do_PUT(self, real=handler_class.do_PUT, count=[0]):
        count[0] += 1
        if count[0] > 1:
            return real(self)

        # Read half the data
        self.rfile.read(min(BUFSIZE, int(self.headers['Content-Length']) // 2))

        # Then close the connection silently
        self.close_connection = True

    monkeypatch.setattr(handler_class, 'do_PUT', do_PUT)

    # Write a big object. We need to write random data, or
    # compression while make the payload too small
    with open('/dev/urandom', 'rb') as rnd:
        buf = rnd.read(5 * BUFSIZE)
    with pytest.raises(ConnectionClosed):
        backend.store('borg', buf)

    enable_temp_fail(backend)
    backend.store('borg', buf)


@pytest.mark.with_backend('gs/aes+zlib', require_mock_server=True)
def test_expired_token_get(backend, monkeypatch):
    '''Test handling of expired OAuth token'''

    key = 'borg'
    data = b'hello there, let us see whats going on'

    # Monkeypatch backend class to check if token is refreshed
    token_refreshed = False

    def _get_access_token(self):
        nonlocal token_refreshed
        token_refreshed = True
        self.access_token[self.password] = 'foobar'

    monkeypatch.setattr(GSBackend, '_get_access_token', _get_access_token)

    # Store some data
    backend[key] = data

    # Monkeypatch request handler to produce error
    handler_class = mock_server.GSRequestHandler

    def do_GET(self, real=handler_class.do_GET, count=[0]):
        count[0] += 1
        if count[0] > 1:
            return real(self)
        else:
            self.send_error(401, code='AuthenticationRequired')

    monkeypatch.setattr(handler_class, 'do_GET', do_GET)

    token_refreshed = False
    assert backend[key] == data
    assert token_refreshed


@pytest.mark.with_backend('gs/aes+zlib', require_mock_server=True)
def test_expired_token_put(backend, monkeypatch):
    '''Test handling of expired OAuth token'''

    key = 'borg'
    data = b'hello there, let us see whats going on'

    # Monkeypatch backend class to check if token is refreshed
    token_refreshed = False

    def _get_access_token(self):
        nonlocal token_refreshed
        token_refreshed = True
        self.access_token[self.password] = 'foobar'

    monkeypatch.setattr(GSBackend, '_get_access_token', _get_access_token)

    # Monkeypatch request handler to produce error
    handler_class = mock_server.GSRequestHandler

    def do_PUT(self, real=handler_class.do_PUT, count=[0]):
        count[0] += 1
        if count[0] > 1:
            return real(self)
        else:
            self.rfile.read(int(self.headers['Content-Length']))
            self.send_error(401, code='AuthenticationRequired')

    monkeypatch.setattr(handler_class, 'do_PUT', do_PUT)

    token_refreshed = False
    backend[key] = data
    assert token_refreshed


@pytest.mark.with_backend('s3c/aes+zlib', require_mock_server=True)
def test_conn_abort(backend, monkeypatch):
    '''Close connection while sending data'''

    data = b'hello there, let us see whats going on'
    key = 'borg'
    backend[key] = data

    # Monkeypatch request handler
    handler_class = mock_server.S3CRequestHandler

    def send_data(self, data, count=[0]):
        count[0] += 1
        if count[0] >= 3:
            self.wfile.write(data)
        else:
            self.wfile.write(data[: len(data) // 2])
            self.close_connection = True

    monkeypatch.setattr(handler_class, 'send_data', send_data)

    with pytest.raises(ConnectionClosed):
        backend.fetch(key)

    # Since we have disabled retry on failure, the connection is not automatically
    # reset after the error.
    backend.backend.conn.reset()

    enable_temp_fail(backend)
    assert backend[key] == data



#!/usr/bin/env python3
'''
t4_authinfo.py - this file is part of S3QL.

Copyright © 2019 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

if __name__ == '__main__':
    import sys

    import pytest

    sys.exit(pytest.main([__file__] + sys.argv[1:]))

import shutil
import subprocess
import tempfile
from argparse import Namespace

import pytest


def test_invalid_option(s3ql_cmd_argv, reg_output):
    with tempfile.NamedTemporaryFile('wt') as fh:
        print('[entry1]', 'storage-url: local:///foo', 'invalid-option: bla', '', file=fh, sep='\n')
        fh.flush()

        proc = subprocess.Popen(
            s3ql_cmd_argv('fsck.s3ql')
            + ['--quiet', '--authfile', fh.name, '--log', 'none', 'local:///foo/bar'],
            stdin=subprocess.PIPE,
            universal_newlines=True,
        )
        proc.stdin.close()
        assert proc.wait() == 2

        reg_output(r"ERROR: '/com' does not exist", count=1)
        proc = subprocess.Popen(
            s3ql_cmd_argv('fsck.s3ql')
            + ['--quiet', '--authfile', fh.name, '--log', 'none', 'local:///com'],
            stdin=subprocess.PIPE,
            universal_newlines=True,
        )
        proc.stdin.close()
        assert proc.wait() == 16


def test_invalid_backend_option(s3ql_cmd_argv):
    with tempfile.NamedTemporaryFile('wt') as fh:
        print(
            '[entry1]',
            'storage-url: local:///foo',
            'backend-options: invalid-key',
            file=fh,
            sep='\n',
        )
        fh.flush()

        proc = subprocess.Popen(
            s3ql_cmd_argv('fsck.s3ql')
            + ['--quiet', '--authfile', fh.name, '--log', 'none', 'local:///foo/bar'],
            stdin=subprocess.PIPE,
            universal_newlines=True,
        )
        proc.stdin.close()
        assert proc.wait() == 3


def test_option_precedence(s3ql_cmd_argv, reg_output):
    with tempfile.NamedTemporaryFile('wt') as fh:
        print(
            '[entry1]',
            'storage-url: s3://',
            'backend-options: invalid-key',
            '',
            '[entry2]',
            'storage-url: s3://foo',
            'backend-options: no-ssl',
            file=fh,
            sep='\n',
        )
        fh.flush()

        reg_output(r"ERROR: Invalid storage URL", count=1)
        proc = subprocess.Popen(
            s3ql_cmd_argv('fsck.s3ql')
            + ['--quiet', '--authfile', fh.name, '--log', 'none', 's3://foo'],
            stdin=subprocess.PIPE,
            universal_newlines=True,
        )
        proc.stdin.close()
        assert proc.wait() == 2


@pytest.fixture()
def context():
    ctx = Namespace()
    ctx.cache_dir = tempfile.mkdtemp(prefix='s3ql-cache-')
    ctx.backend_dir = tempfile.mkdtemp(prefix='s3ql-backend-')
    ctx.storage_url = 'local://' + ctx.backend_dir

    yield ctx

    shutil.rmtree(ctx.cache_dir)
    shutil.rmtree(ctx.backend_dir)


def test_passphrase(context, reg_output, s3ql_cmd_argv):
    passphrase = 'out3d'
    proc = subprocess.Popen(
        s3ql_cmd_argv('mkfs.s3ql')
        + [
            '-L',
            'test fs',
            '--data-block-size',
            '500',
            '--authfile',
            '/dev/null',
            '--cachedir',
            context.cache_dir,
            '--quiet',
            context.storage_url,
        ],
        stdin=subprocess.PIPE,
        universal_newlines=True,
    )

    print(passphrase, file=proc.stdin)
    print(passphrase, file=proc.stdin)
    proc.stdin.close()
    assert proc.wait() == 0

    with tempfile.NamedTemporaryFile('wt') as fh:
        print(
            '[entry1]',
            'storage-url: local://',
            'fs-passphrase: clearly wrong',
            '',
            '[entry2]',
            'storage-url: %s' % context.storage_url,
            'fs-passphrase: %s' % passphrase,
            file=fh,
            sep='\n',
        )
        fh.flush()

        proc = subprocess.Popen(
            s3ql_cmd_argv('fsck.s3ql')
            + [
                '--quiet',
                '--authfile',
                fh.name,
                '--cachedir',
                context.cache_dir,
                '--log',
                'none',
                context.storage_url,
            ],
            stdin=subprocess.PIPE,
            universal_newlines=True,
        )
        proc.stdin.close()
        assert proc.wait() == 0



'''
multi_lock.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import logging

import trio

try:
    from contextlib import asynccontextmanager
except ModuleNotFoundError:
    from async_generator import asynccontextmanager

__all__ = ["MultiLock"]

log = logging.getLogger(__name__)


class MultiLock:
    """Provides locking for multiple objects.

    This class provides locking for a dynamically changing set of objects: The
    `acquire` and `release` methods have an additional argument, the locking
    key. Only locks with the same key can actually see each other, so that
    several threads can hold locks with different locking keys at the same time.

    MultiLock instances can be used as context managers.

    Note that it is actually possible for one task to release a lock that has
    been obtained by a different thread. This is not a bug but a feature.
    """

    def __init__(self):
        self.locked_keys = set()
        self.cond = trio.Condition()

    @asynccontextmanager
    async def __call__(self, *key):
        await self.acquire(*key)
        try:
            yield
        finally:
            await self.release(*key)

    def acquire_nowait(self, *key):
        '''Try to acquire lock for given key.

        Return True on success, False on failure.
        '''

        try:
            self.cond.acquire_nowait()
        except trio.WouldBlock:
            return False

        try:
            if key in self.locked_keys:
                return False
            else:
                self.locked_keys.add(key)
                return True
        finally:
            self.cond.release()

    async def acquire(self, *key):
        '''Acquire lock for given key.'''

        async with self.cond:
            while key in self.locked_keys:
                await self.cond.wait()

            self.locked_keys.add(key)

    async def release(self, *key, noerror=False):
        """Release lock on given key

        If noerror is False, do not raise exception if *key* is
        not locked.
        """

        async with self.cond:
            if noerror:
                self.locked_keys.discard(key)
            else:
                self.locked_keys.remove(key)
            self.cond.notify_all()



'''
conftest.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.


This module is loaded automatically by py.test and is used to
initialize logging and adjust the load path before running
any tests.
'''

import faulthandler
import gc
import logging.handlers
import os.path
import signal
import sys
import time

import pytest
import pytest_trio

assert pytest_trio  # suppress unused import warning


# If a test fails, wait a moment before retrieving the captured
# stdout/stderr. When using a server process (like in t4_fuse.py), this makes
# sure that we capture any potential output of the server that comes *after* a
# test has failed. For example, if a request handler raises an exception, the
# server first signals an error to FUSE (causing the test to fail), and then
# logs the exception. Without the extra delay, the exception will go into
# nowhere.
@pytest.hookimpl(hookwrapper=True)
def pytest_pyfunc_call(pyfuncitem):
    outcome = yield
    failed = outcome.excinfo is not None
    if failed:
        time.sleep(1)


@pytest.fixture(scope="class")
def s3ql_cmd_argv(request):
    '''Provide argument list to execute s3ql commands in tests'''

    if request.config.getoption('installed'):
        yield lambda cmd: [cmd]
    else:
        basedir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
        yield lambda cmd: [sys.executable, os.path.join(basedir, 'bin', cmd)]


# Enable output checks
pytest_plugins = ('pytest_checklogs',)


@pytest.fixture()
def pass_reg_output(request, reg_output):
    '''Provide reg_output function to UnitTest instances'''
    request.instance.reg_output = reg_output


@pytest.fixture()
def pass_s3ql_cmd_argv(request, s3ql_cmd_argv):
    '''Provide s3ql_cmd_argv function to UnitTest instances'''
    request.instance.s3ql_cmd_argv = s3ql_cmd_argv


def pytest_addoption(parser):
    group = parser.getgroup("terminal reporting")
    group._addoption(
        "--logdebug",
        action="append",
        metavar='<module>',
        help="Activate debugging output from <module> for tests. Use `all` "
        "to get debug messages from all modules. This option can be "
        "specified multiple times.",
    )

    group = parser.getgroup("general")
    group._addoption(
        "--installed", action="store_true", default=False, help="Test the installed package."
    )


def pytest_configure(config):
    # If we are running from the S3QL source directory, make sure that we
    # load modules from here
    basedir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    if not config.getoption('installed'):
        if os.path.exists(os.path.join(basedir, 'setup.py')) and os.path.exists(
            os.path.join(basedir, 'src', 's3ql', '__init__.py')
        ):
            sys.path = [os.path.join(basedir, 'src')] + sys.path

    # Enable all warnings for subprocesses. This is not optimal, because they will be treated like
    # other log messages (rather than captured as warnings by pytest) and can thus trigger test
    # failures. Therefore, we only enable this when running from Git. Note that logging.py
    # has additional logic to suppress some unactionable warnings, which is triggered by the
    # dummy entry.
    if os.path.exists(os.path.join(basedir, '.git')):
        os.environ['PYTHONWARNINGS'] = 'default,default:set_by_conftest'

    # Enable faulthandler
    faultlog_fd = os.open(
        os.path.join(basedir, 'tests', 'test_crit.log'),
        flags=os.O_APPEND | os.O_CREAT | os.O_WRONLY,
        mode=0o644,
    )
    faulthandler.enable(faultlog_fd)
    faulthandler.register(signal.SIGUSR1, file=faultlog_fd)

    # Configure logging. We don't set a default handler but rely on
    # the catchlog pytest plugin.
    logdebug = config.getoption('logdebug')
    root_logger = logging.getLogger()
    if logdebug is not None:
        if 'all' in logdebug:
            root_logger.setLevel(logging.DEBUG)
        else:
            for module in logdebug:
                logging.getLogger(module).setLevel(logging.DEBUG)
    else:
        root_logger.setLevel(logging.INFO)


# Run gc.collect() at the end of every test, so that we get ResourceWarnings
# as early as possible.
def pytest_runtest_teardown(item, nextitem):
    gc.collect()



'''
mount.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import _thread
import argparse
import atexit
import faulthandler
import functools
import logging
import os
import platform
import re
import resource
import shutil
import signal
import subprocess
import sys
import threading
import time
from contextlib import AsyncExitStack
from typing import Tuple

import pyfuse3
import trio

from . import fs
from .backends.pool import BackendPool
from .block_cache import BlockCache
from .common import get_backend_factory, is_mounted
from .daemonize import daemonize
from .database import (
    Connection,
    FsAttributes,
    download_metadata,
    expire_objects,
    read_cached_params,
    read_remote_params,
    upload_metadata,
    upload_params,
    write_params,
)
from .inode_cache import InodeCache
from .logging import QuietError, setup_logging, setup_warnings
from .parse_args import ArgumentParser

log = logging.getLogger(__name__)


def install_thread_excepthook():
    """work around sys.excepthook thread bug

    See http://bugs.python.org/issue1230540.

    Call once from __main__ before creating any threads. If using
    psyco, call psyco.cannotcompile(threading.Thread.run) since this
    replaces a new-style class method.
    """

    init_old = threading.Thread.__init__

    def init(self, *args, **kwargs):
        init_old(self, *args, **kwargs)
        run_old = self.run

        def run_with_except_hook(*args, **kw):
            try:
                run_old(*args, **kw)
            except SystemExit:
                raise
            except:
                sys.excepthook(*sys.exc_info())

        self.run = run_with_except_hook

    threading.Thread.__init__ = init


def main(args=None):
    '''Mount S3QL file system'''

    if args is None:
        args = sys.argv[1:]

    install_thread_excepthook()
    setup_warnings()
    options = parse_args(args)

    # Save handler so that we can remove it when daemonizing
    stdout_log_handler = setup_logging(options)

    if not os.path.exists(options.mountpoint):
        raise QuietError('Mountpoint does not exist.', exitcode=36)

    # Check if fs is mounted on this computer
    # This is not foolproof but should prevent common mistakes
    if is_mounted(options.storage_url):
        raise QuietError('File system already mounted elsewhere on this machine.', exitcode=40)

    if options.threads is None:
        options.threads = determine_threads(options)

    avail_fd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
    if avail_fd == resource.RLIM_INFINITY:
        avail_fd = 4096
    resource.setrlimit(resource.RLIMIT_NOFILE, (avail_fd, avail_fd))

    # Subtract some fd's for random things we forgot, and a fixed number for
    # each upload thread (because each thread is using at least one socket and
    # at least one temporary file)
    avail_fd -= 32 + 3 * options.threads

    if options.max_cache_entries is None:
        if avail_fd <= 64:
            raise QuietError("Not enough available file descriptors.", exitcode=37)
        log.info('Autodetected %d file descriptors available for cache entries', avail_fd)
        options.max_cache_entries = avail_fd
    else:
        if options.max_cache_entries > avail_fd:
            log.warning(
                "Up to %d cache entries requested, but detected only %d "
                "available file descriptors.",
                options.max_cache_entries,
                avail_fd,
            )
            options.max_cache_entries = avail_fd

    if options.profile:
        import cProfile
        import pstats

        prof = cProfile.Profile()
        prof.runcall(trio.run, main_async, options, stdout_log_handler)
        with open('s3ql_profile.txt', 'w') as fh:
            prof.dump_stats('s3ql_profile.dat')
            p = pstats.Stats('s3ql_profile.dat', stream=fh)
            p.strip_dirs()
            p.sort_stats('cumulative')
            p.print_stats(50)
            p.sort_stats('time')
            p.print_stats(50)
    else:
        # trio.run(main_async, options, stdout_log_handler,
        #         instruments=[Tracer()])
        trio.run(main_async, options, stdout_log_handler)


class Tracer(trio.abc.Instrument):
    def _print_with_task(self, msg, task):
        print("{}: {}".format(msg, task.name))

    def task_spawned(self, task):
        self._print_with_task("### new task spawned", task)

    def task_scheduled(self, task):
        self._print_with_task("### task scheduled", task)

    def before_task_step(self, task):
        self._print_with_task(">>> about to run one step of task", task)

    def after_task_step(self, task):
        self._print_with_task("<<< task step finished", task)

    def task_exited(self, task):
        self._print_with_task("### task exited", task)

    def before_io_wait(self, timeout):
        if timeout:
            print("### waiting for I/O for up to {} seconds".format(timeout))
        else:
            print("### doing a quick check for I/O")
        self._sleep_time = trio.current_time()

    def after_io_wait(self, timeout):
        duration = trio.current_time() - self._sleep_time
        print("### finished I/O check (took {} seconds)".format(duration))


async def main_async(options, stdout_log_handler):
    # Get paths
    cachepath = options.cachepath

    backend_factory = get_backend_factory(options)
    backend_pool = BackendPool(backend_factory)
    atexit.register(backend_pool.flush)

    with backend_pool() as backend:
        (param, db) = get_metadata(backend, cachepath)

    # Handle --cachesize
    rec_cachesize = options.max_cache_entries * param.data_block_size / 2
    avail_cache = shutil.disk_usage(os.path.dirname(cachepath))[2] / 1024
    if options.cachesize is None:
        options.cachesize = min(rec_cachesize, 0.8 * avail_cache)
        log.info('Setting cache size to %d MB', options.cachesize / 1024)
    elif options.cachesize > avail_cache:
        log.warning(
            'Requested cache size %d MB, but only %d MB available',
            options.cachesize / 1024,
            avail_cache / 1024,
        )

    if options.nfs:
        # NFS may try to look up '..', so we have to speed up this kind of query
        log.info('Creating NFS indices...')
        db.execute('CREATE INDEX IF NOT EXISTS ix_contents_inode ON contents(inode)')

    else:
        db.execute('DROP INDEX IF EXISTS ix_contents_inode')

    metadata_upload_task = MetadataUploadTask(param, backend_pool, db, options)

    async with trio.open_nursery() as nursery:
        async with AsyncExitStack() as cm:
            block_cache = BlockCache(
                backend_pool,
                db,
                cachepath + '-cache',
                options.cachesize * 1024,
                options.max_cache_entries,
            )
            cm.push_async_callback(block_cache.destroy, options.keep_cache)

            operations = fs.Operations(
                block_cache,
                db,
                max_obj_size=param.data_block_size,
                inode_cache=InodeCache(db, param.inode_gen),
                upload_task=metadata_upload_task,
            )
            cm.push_async_callback(operations.destroy)
            block_cache.fs = operations
            metadata_upload_task.fs = operations

            log.info('Mounting %s at %s...', options.storage_url, options.mountpoint)
            try:
                pyfuse3.init(operations, options.mountpoint, get_fuse_opts(options))
            except RuntimeError as exc:
                raise QuietError(str(exc), exitcode=39)

            unmount_clean = False

            def unmount():
                log.info("Unmounting file system...")
                pyfuse3.close(unmount=unmount_clean)

            cm.callback(unmount)

            if options.fg or options.systemd:
                faulthandler.enable()
                faulthandler.register(signal.SIGUSR1)
            else:
                if stdout_log_handler:
                    logging.getLogger().removeHandler(stdout_log_handler)
                crit_log_fd = os.open(
                    os.path.join(options.cachedir, 'mount.s3ql_crit.log'),
                    flags=os.O_APPEND | os.O_CREAT | os.O_WRONLY,
                    mode=0o644,
                )
                faulthandler.enable(crit_log_fd)
                faulthandler.register(signal.SIGUSR1, file=crit_log_fd)
                daemonize(options.cachedir)

            param.seq_no += 1
            param.is_mounted = True
            write_params(cachepath, param)
            upload_params(backend, param)

            block_cache.init(options.threads)

            nursery.start_soon(metadata_upload_task.run, name='metadata-upload-task')
            cm.callback(metadata_upload_task.stop)

            commit_task = CommitTask(block_cache, options.dirty_block_upload_delay)
            nursery.start_soon(commit_task.run, name='commit-task')
            cm.callback(commit_task.stop)

            exc_info = setup_exchook()

            received_signals = []
            old_handler = None

            def signal_handler(signum, frame, _main_thread=_thread.get_ident()):
                log.debug('Signal %d received', signum)
                if pyfuse3.trio_token is None:
                    log.debug('FUSE loop not running, calling parent handler...')
                    return old_handler(signum, frame)
                log.debug('Calling pyfuse3.terminate()...')
                if _thread.get_ident() == _main_thread:
                    pyfuse3.terminate()
                else:
                    trio.from_thread.run_sync(pyfuse3.terminate, trio_token=pyfuse3.trio_token)
                received_signals.append(signum)

            signal.signal(signal.SIGTERM, signal_handler)
            old_handler = signal.signal(signal.SIGINT, signal_handler)

            if options.systemd:
                import systemd.daemon

                systemd.daemon.notify('READY=1')

            await pyfuse3.main()

            if received_signals == [signal.SIGINT]:
                log.info('SIGINT received, exiting normally.')
            elif received_signals:
                log.info('Received termination signal, exiting without data upload.')
                raise RuntimeError('Received signal(s) %s, terminating' % (received_signals,))

            # Re-raise if main loop terminated due to exception in other thread
            if exc_info:
                (exc_inst, exc_tb) = exc_info
                raise exc_inst.with_traceback(exc_tb)

            log.info("FUSE main loop terminated.")

            unmount_clean = True

    # At this point, there should be no other threads left
    param.is_mounted = False
    db.close()

    # Do not update .params yet, dump_metadata() may fail if the database is
    # corrupted, in which case we want to force an fsck.
    if operations.failsafe:
        log.warning('File system errors encountered, marking for fsck.')
        param.needs_fsck = True

    with backend_pool() as backend:
        param.last_modified = time.time()
        upload_metadata(backend, db, param)
        write_params(cachepath, param)
        upload_params(backend, param)
        expire_objects(backend)

    log.info('All done.')


def get_system_memory():
    '''Attempt to determine total system memory

    If amount cannot be determined, emits warning and
    returns -1.
    '''

    # MacOS X doesn't support sysconf('SC_PHYS_PAGES')
    if platform.system() == 'Darwin':
        try:
            out = subprocess.check_output(['sysctl', 'hw.memsize'], universal_newlines=True)
        except subprocess.CalledProcessError as exc:
            log.warning('Cannot determine system memory, sysctl failed with %s', exc.output)
            return -1

        # output of sysctl is 'hw.memsize: #'. Strip the prefix.
        hit = re.match(r'^hw.memsize: ([0-9]+)$', out)
        if not hit:
            log.warning('Cannot determine system memory, unable to parse sysctl output.')
            return -1

        return int(hit.group(1))

    else:
        try:
            return os.sysconf('SC_PHYS_PAGES') * os.sysconf('SC_PAGESIZE')
        except ValueError:
            log.warning('Unable to determine number of CPU cores (sysconf failed).')
            return -1


# Memory required for LZMA compression in MB (from xz(1))
LZMA_MEMORY = {0: 3, 1: 9, 2: 17, 3: 32, 4: 48, 5: 94, 6: 94, 7: 186, 8: 370, 9: 674}


def determine_threads(options):
    '''Return optimum number of upload threads'''

    try:
        cores = os.sysconf('SC_NPROCESSORS_ONLN')
    except ValueError:
        log.warning('Unable to determine number of CPU cores (sysconf failed).')
        cores = -1

    memory = get_system_memory()

    if options.compress[0] == 'lzma':
        # Keep this in sync with compression level in backends/common.py
        # Memory usage according to man xz(1)
        mem_per_thread = LZMA_MEMORY[options.compress[1]] * 1024**2
    else:
        # Only check LZMA memory usage
        mem_per_thread = 0

    if cores == -1:
        log.warning("Can't determine number of cores, using 2 upload threads.")
        return 2
    elif memory == -1 and mem_per_thread != 0:
        log.warning("Can't determine available memory, using 2 upload threads.")
        return 2
    elif 2 * cores * mem_per_thread > (memory / 2):
        threads = min(int((memory / 2) // mem_per_thread), 10)
        if threads > 0:
            log.info('Using %d upload threads (memory limited).', threads)
        else:
            log.warning(
                'Compression will require %d MiB memory (%d%% of total system memory',
                mem_per_thread / 1024**2,
                mem_per_thread * 100 / memory,
            )
            threads = 1
        return threads
    else:
        threads = min(2 * cores, 10)
        log.info("Using %d upload threads.", threads)
        return threads


def get_metadata(backend, cachepath) -> Tuple[FsAttributes, Connection]:
    '''Retrieve metadata'''

    db = None
    param = read_remote_params(backend)
    local_param = read_cached_params(cachepath, min_seq=param.seq_no)
    if local_param is not None:
        if local_param.seq_no > param.seq_no:
            raise QuietError("File system not unmounted cleanly, run fsck!", exitcode=30)
        log.info('Using cached metadata.')
        assert local_param == param
        db = Connection(cachepath + '.db', param.metadata_block_size)

    if param.is_mounted:
        raise QuietError(
            'Backend reports that fs is still mounted elsewhere, aborting.', exitcode=31
        )

    if param.needs_fsck:
        raise QuietError("File system damaged or not unmounted cleanly, run fsck!", exitcode=30)
    if time.time() - param.last_fsck > 60 * 60 * 24 * 31:
        log.warning(
            'Last file system check was more than 1 month ago, running fsck.s3ql is recommended.'
        )

    # Download metadata
    if not db:
        log.info('Downloading metadata...')
        db = download_metadata(backend, cachepath + '.db', param)

        # Drop cache
        if os.path.exists(cachepath + '-cache'):
            shutil.rmtree(cachepath + '-cache')

    return (param, db)


def get_fuse_opts(options):
    '''Return fuse options for given command line options'''

    fsname = options.fs_name
    if not fsname:
        fsname = options.storage_url
    fuse_opts = ['fsname=%s' % fsname, 'subtype=s3ql']

    if options.allow_other:
        fuse_opts.append('allow_other')
    if options.allow_root:
        fuse_opts.append('allow_root')
    if options.allow_other or options.allow_root:
        fuse_opts.append('default_permissions')

    return fuse_opts


def parse_args(args):
    '''Parse command line'''

    # Parse fstab-style -o options
    if '--' in args:
        max_idx = args.index('--')
    else:
        max_idx = len(args)
    if '-o' in args[:max_idx]:
        pos = args.index('-o')
        val = args[pos + 1]
        del args[pos]
        del args[pos]
        for opt in reversed(val.split(',')):
            if '=' in opt:
                (key, val) = opt.split('=')
                args.insert(pos, val)
                args.insert(pos, '--' + key)
            else:
                if opt in (
                    'rw',
                    'defaults',
                    'auto',
                    'noauto',
                    'user',
                    'nouser',
                    'dev',
                    'nodev',
                    'suid',
                    'nosuid',
                    'atime',
                    'diratime',
                    'exec',
                    'noexec',
                    'group',
                    'mand',
                    'nomand',
                    '_netdev',
                    'nofail',
                    'norelatime',
                    'strictatime',
                    'owner',
                    'users',
                    'nobootwait',
                ):
                    continue
                elif opt == 'ro':
                    raise QuietError('Read-only mounting not supported.', exitcode=35)
                args.insert(pos, '--' + opt)

    parser = ArgumentParser(description="Mount an S3QL file system.")

    parser.add_log('~/.s3ql/mount.log')
    parser.add_cachedir()
    parser.add_debug()
    parser.add_quiet()
    parser.add_backend_options()
    parser.add_version()
    parser.add_storage_url()
    parser.add_compress()

    parser.add_argument(
        "mountpoint",
        metavar='<mountpoint>',
        type=os.path.abspath,
        help='Where to mount the file system',
    )
    parser.add_argument(
        "--cachesize",
        type=int,
        default=None,
        metavar='<size>',
        help="Cache size in KiB (default: autodetect).",
    )
    parser.add_argument(
        "--max-cache-entries",
        type=int,
        default=None,
        metavar='<num>',
        help="Maximum number of entries in cache (default: autodetect). "
        'Each cache entry requires one file descriptor, so if you increase '
        'this number you have to make sure that your process file descriptor '
        'limit (as set with `ulimit -n`) is high enough (at least the number '
        'of cache entries + 100).',
    )
    parser.add_argument(
        "--keep-cache",
        action="store_true",
        default=False,
        help="Do not purge locally cached files on exit.",
    )
    parser.add_argument(
        "--allow-other",
        action="store_true",
        default=False,
        help='Normally, only the user who called `mount.s3ql` can access the mount '
        'point. This user then also has full access to it, independent of '
        'individual file permissions. If the `--allow-other` option is '
        'specified, other users can access the mount point as well and '
        'individual file permissions are taken into account for all users.',
    )
    parser.add_argument(
        "--allow-root",
        action="store_true",
        default=False,
        help='Like `--allow-other`, but restrict access to the mounting user and the root user.',
    )
    parser.add_argument(
        "--dirty-block-upload-delay",
        action="store",
        type=int,
        default=10,
        metavar='<seconds>',
        help="Upload delay for dirty blocks in seconds (default: 10 seconds).",
    )
    parser.add_argument(
        "--fg", action="store_true", default=False, help="Do not daemonize, stay in foreground"
    )
    parser.add_argument(
        "--fs-name",
        default=None,
        help="Mount name passed to fuse, the name will be shown in the first "
        "column of the system mount command output. If not specified your "
        "storage url is used.",
    )
    parser.add_argument(
        "--systemd",
        action="store_true",
        default=False,
        help="Run as systemd unit. Consider specifying --log none as well "
        "to make use of journald.",
    )
    parser.add_argument(
        "--metadata-backup-interval",
        action="store",
        type=int,
        default=6 * 60 * 60,
        metavar='<seconds>',
        help='Interval between metadata backups. Should the filesystem crash while mounted, '
        'modifications made after the most recent metadata backup may be lost. During '
        'backups, the filesystem will be unresponsile. Default: 6h',
    )
    parser.add_argument(
        "--threads",
        action="store",
        type=int,
        default=None,
        metavar='<no>',
        help='Number of parallel upload threads to use (default: auto).',
    )
    parser.add_argument(
        "--nfs",
        action="store_true",
        default=False,
        help='Enable some optimizations for exporting the file system '
        'over NFS. (default: %(default)s)',
    )
    parser.add_argument("--profile", action="store_true", default=False, help=argparse.SUPPRESS)

    options = parser.parse_args(args)

    if options.allow_other and options.allow_root:
        parser.error("--allow-other and --allow-root are mutually exclusive.")

    if not options.log and not (options.fg or options.systemd):
        parser.error("Please activate logging to a file or syslog, or use the --fg option.")

    if options.metadata_backup_interval == 0:
        options.metadata_backup_interval = None

    return options


class MetadataUploadTask:
    '''
    Periodically upload metadata. Upload is done every `interval`
    seconds, and whenever `event` is set. To terminate thread,
    set `quit` attribute as well as `event` event.
    '''

    def __init__(self, params: FsAttributes, backend_pool, db, options):
        super().__init__()
        self.backend_pool = backend_pool
        self.db = db
        self.event = trio.Event()
        self.quit = False
        self.params = params
        self.options = options

    async def run(self):
        log.debug('started')

        while not self.quit:
            with trio.move_on_after(self.options.metadata_backup_interval):
                await self.event.wait()

            if self.quit:
                break

            self.event = trio.Event()  # reset

            with self.backend_pool() as backend:
                # Upload asynchronously twice to reduce the amount of data left for
                # synchronous upload.
                self.db.checkpoint()
                for _ in range(2):
                    await trio.to_thread.run_sync(
                        functools.partial(
                            upload_metadata,
                            backend,
                            self.db,
                            self.params,
                            update_params=False,
                            incremental=True,
                        )
                    )

                # Now upload synchronously to get consistent snapshot (at the cost of stopping file
                # system operation). As a future optimization, we could first copy all modified
                # blocks locally, and then upload async...
                self.db.checkpoint()
                self.params.last_modified = time.time()
                upload_metadata(
                    backend,
                    self.db,
                    self.params,
                    update_params=True,
                    incremental=True,
                )
                write_params(self.options.cachepath, self.params)
                upload_params(backend, self.params)

                # Write a new params file immediately, so that we're in the same state as right
                # after mounting and there is no window where we could have metadata_* objects with
                # a sequence number for which there is no corresponding s3ql_params_* object.
                self.params.seq_no += 1
                write_params(self.options.cachepath, self.params)
                upload_params(backend, self.params)

                await trio.to_thread.run_sync(expire_objects, backend)

        log.debug('finished')

    def stop(self):
        '''Signal thread to terminate'''

        log.debug('started')
        self.quit = True
        self.event.set()


def setup_exchook():
    '''Terminate FUSE main loop if any thread terminates with an exception

    The exc_info will be saved in the list object returned by this function.
    '''

    main_thread = _thread.get_ident()
    old_exchook = sys.excepthook
    exc_info = []

    def exchook(exc_type, exc_inst, tb):
        reporting_thread = _thread.get_ident()
        if reporting_thread != main_thread:
            if exc_info:
                log.warning("Unhandled top-level exception during shutdown (will not be re-raised)")
            else:
                log.error("Unhandled exception in thread, terminating", exc_info=True)
                exc_info.append(exc_inst)
                exc_info.append(tb)
                trio.from_thread.run_sync(pyfuse3.terminate, trio_token=pyfuse3.trio_token)

            old_exchook(exc_type, exc_inst, tb)

        # If the main thread re-raised exception, there is no need to call
        # excepthook again
        elif exc_info and exc_info[0] is exc_inst:
            log.debug('Suppressing exception hook for re-raised %s', exc_inst)
        else:
            old_exchook(exc_type, exc_inst, tb)

    sys.excepthook = exchook

    return exc_info


class CommitTask:
    '''
    Periodically upload dirty blocks.
    '''

    def __init__(self, block_cache, dirty_block_upload_delay):
        super().__init__()
        self.block_cache = block_cache
        self.stop_event = trio.Event()
        self.dirty_block_upload_delay = dirty_block_upload_delay

    async def run(self):
        log.debug('started')

        while not self.stop_event.is_set():
            did_sth = False

            stamp = time.time()
            # Need to make copy, since we aren't allowed to change
            # dict while iterating through it. The performance hit doesn't seem
            # to be that bad:
            # >>> from timeit import timeit
            # >>> timeit("k=0\nfor el in list(d.values()):\n k += el",
            # ... setup='\nfrom collections import OrderedDict\nd = OrderedDict()\nfor i in range(5000):\n d[i]=i\n',
            # ... number=500)/500 * 1e3
            # 1.3769531380003173
            # >>> timeit("k=0\nfor el in d.values(n:\n k += el",
            # ... setup='\nfrom collections import OrderedDict\nd = OrderedDict()\nfor i in range(5000):\n d[i]=i\n',
            # ... number=500)/500 * 1e3
            # 1.456586996000624
            for el in list(self.block_cache.cache.values()):
                if (
                    self.stop_event.is_set()
                    or stamp - el.last_write < self.dirty_block_upload_delay
                ):
                    break
                if el.dirty and el not in self.block_cache.in_transit:
                    await self.block_cache.upload_if_dirty(el)
                    did_sth = True

            if not did_sth:
                with trio.move_on_after(5):
                    await self.stop_event.wait()

        log.debug('finished')

    def stop(self):
        '''Signal thread to terminate'''

        log.debug('started')
        self.stop_event.set()


if __name__ == '__main__':
    main(sys.argv[1:])



'''
s3.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import logging
import re
from xml.sax.saxutils import escape as xml_escape

from ..logging import QuietError
from . import s3c4
from .common import retry
from .s3c import get_S3Error

log = logging.getLogger(__name__)

# Maximum number of keys that can be deleted at once
MAX_KEYS = 1000

# Pylint goes berserk with false positives
# pylint: disable=E1002,E1101


class Backend(s3c4.Backend):
    """A backend to store data in Amazon S3

    This class uses standard HTTP connections to connect to S3.
    """

    known_options = (s3c4.Backend.known_options | {'sse', 'rrs', 'ia', 'oia', 'it'}) - {
        'dumb-copy',
        'disable-expect100',
        'sig-region',
    }

    def __init__(self, options):
        self.region = None
        super().__init__(options)
        self._set_storage_options(self._extra_put_headers)
        self.sig_region = self.region

    def _parse_storage_url(self, storage_url, ssl_context):
        hit = re.match(r'^s3s?://([^/]+)/([^/]+)(?:/(.*))?$', storage_url)
        if not hit:
            raise QuietError('Invalid storage URL', exitcode=2)

        self.region = hit.group(1)
        bucket_name = hit.group(2)

        # http://docs.amazonwebservices.com/AmazonS3/2006-03-01/dev/BucketRestrictions.html
        if not re.match('^[a-z0-9][a-z0-9.-]{1,60}[a-z0-9]$', bucket_name):
            raise QuietError('Invalid bucket name.', exitcode=2)

        if self.region.startswith('cn-'):
            hostname = 's3.%s.amazonaws.com.cn' % self.region
        else:
            hostname = 's3.%s.amazonaws.com' % self.region

        prefix = hit.group(3) or ''
        port = 443 if ssl_context else 80
        return (hostname, port, bucket_name, prefix)

    def __str__(self):
        return 'Amazon S3 bucket %s, prefix %s' % (self.bucket_name, self.prefix)

    @property
    def has_delete_multi(self):
        return True

    def delete_multi(self, keys):
        log.debug('started with %s', keys)

        while len(keys) > 0:
            tmp = keys[:MAX_KEYS]
            try:
                self._delete_multi(tmp)
            finally:
                keys[:MAX_KEYS] = tmp

    def _set_storage_options(self, headers):
        if 'sse' in self.options:
            headers['x-amz-server-side-encryption'] = 'AES256'

        if 'ia' in self.options:
            sc = 'STANDARD_IA'
        elif 'oia' in self.options:
            sc = 'ONEZONE_IA'
        elif 'rrs' in self.options:
            sc = 'REDUCED_REDUNDANCY'
        elif 'it' in self.options:
            sc = 'INTELLIGENT_TIERING'
        else:
            sc = 'STANDARD'
        headers['x-amz-storage-class'] = sc

    @retry
    def _delete_multi(self, keys):
        body = ['<Delete>']
        esc_prefix = xml_escape(self.prefix)
        for key in keys:
            body.append('<Object><Key>%s%s</Key></Object>' % (esc_prefix, xml_escape(key)))
        body.append('</Delete>')
        body = '\n'.join(body).encode('utf-8')
        headers = {'content-type': 'text/xml; charset=utf-8'}

        resp = self._do_request('POST', '/', subres='delete', body=body, headers=headers)
        try:
            root = self._parse_xml_response(resp)
            ns_p = self.xml_ns_prefix

            error_tags = root.findall(ns_p + 'Error')
            if not error_tags:
                # No errors occurred, everything has been deleted
                del keys[:]
                return

            # Some errors occurred, so we need to determine what has
            # been deleted and what hasn't
            offset = len(self.prefix)
            for tag in root.findall(ns_p + 'Deleted'):
                fullkey = tag.find(ns_p + 'Key').text
                assert fullkey.startswith(self.prefix)
                keys.remove(fullkey[offset:])

            if log.isEnabledFor(logging.DEBUG):
                for errtag in error_tags:
                    log.debug(
                        'Delete %s failed with %s',
                        errtag.findtext(ns_p + 'Key')[offset:],
                        errtag.findtext(ns_p + 'Code'),
                    )

            errcode = error_tags[0].findtext(ns_p + 'Code')
            errmsg = error_tags[0].findtext(ns_p + 'Message')
            errkey = error_tags[0].findtext(ns_p + 'Key')[offset:]

            if errcode == 'NoSuchKeyError':
                pass
            else:
                raise get_S3Error(errcode, 'Error deleting %s: %s' % (errkey, errmsg))

        except:
            self.conn.discard()



#!/usr/bin/env python3

import os
import sys

# Add S3QL sources
basedir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, os.path.join(basedir, 'src'))
sys.path.insert(0, os.path.join(basedir, 'util'))
try:
    import s3ql
except ModuleNotFoundError:
    print(
        'ERROR: s3ql python module not found.',
        'Did you run `setup.py build_ext --inplace`?',
        sep='\n',
        file=sys.stderr,
    )
    sys.exit(8)


# Add a custom role for command line options that does not try to reference anything.
def add_literal_role(rolename):
    from docutils import nodes
    from docutils.parsers.rst import roles

    nodeclass = nodes.literal
    generic = roles.GenericRole(rolename, nodeclass)
    role = roles.CustomRole(rolename, generic, {'classes': [rolename]})
    roles.register_local_role(rolename, role)


add_literal_role('cmdopt')
add_literal_role('var')  # A variable defined in a :samp: role

# Add our own Pygments Lexer
import pygments.lexers._mapping as pmap

pmap.LEXERS['CommandLineLexer'] = (
    'cmdline_lexer',
    'CommandLine',
    ('CommandLine', 'commandline'),
    (),
    (),
)


# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.ifconfig', 'sphinx_pipeinclude']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
source_encoding = 'utf-8'

# The master toctree document.
master_doc = 'index'

# Warn about unresolved references
nitpicky = True

# General information about the project.
project = u'S3QL'
copyright = u'© 2008 Nikolaus Rath <Nikolaus@rath.org>'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = s3ql.VERSION
# The full version, including alpha/beta/rc tags.
release = s3ql.RELEASE

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of documents that shouldn't be included in the build.
# unused_docs = []

# List of directories, relative to source directory, that shouldn't be searched
# for source files.
exclude_trees = ['include']

# The reST default role (used for this markup: `text`) to use for all documents.
default_role = 'file'

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'colorful'
highlight_language = 'commandline'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = 'sphinxdoc'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {'stickysidebar': 'true'}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
html_use_modindex = False

# If false, no index is generated.
html_use_index = False

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = ''

# Output file base name for HTML help builder.
htmlhelp_basename = 'S3QLdoc'


# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    ('index', 'manual.tex', u'S3QL Documentation', u'Nikolaus Rath', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_use_modindex = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('man/mkfs', 'mkfs.s3ql', u'Create an S3QL file system', None, 1),
    ('man/mount', 'mount.s3ql', u'Mount an S3QL file system', None, 1),
    ('man/umount', 'umount.s3ql', u'Unmount an S3QL file system', None, 1),
    ('man/fsck', 'fsck.s3ql', u'Check an S3QL file system for errors', None, 1),
    ('man/ctrl', 's3qlctrl', u'Control a mounted S3QL file system', None, 1),
    ('man/adm', 's3qladm', u'Manage S3QL file systems', None, 1),
    ('man/cp', 's3qlcp', u'Copy-on-write replication on S3QL file systems', None, 1),
    ('man/lock', 's3qllock', u'Make trees on an S3QL file system immutable', None, 1),
    ('man/rm', 's3qlrm', u'Fast tree removal on S3QL file systems', None, 1),
    ('man/stat', 's3qlstat', u'Gather S3QL file system statistics', None, 1),
    ('man/expire_backups', 'expire_backups', u'Intelligently expire old backups', None, 1),
    ('man/pcp', 'pcp', u'Recursive, parallel copy of directory trees', None, 1),
    ('man/oauth_client', 's3ql_oauth_client', u'Obtain Google Storage OAuth2 tokens', None, 1),
    ('man/verify', 's3ql_verify', u'Verify data in an S3QL file system', None, 1),
]
